import sys, io
import os
import threading
import time
from pathlib import Path
from collections import OrderedDict
import logging
import psutil
import pprint
import enum

from qtpy import QtCore, QtGui, QtWidgets

from qtpy.QtCore import (QFile, QFileInfo, QPoint, QSettings, QSignalMapper, QTimer,
        QSize, QTextStream, Qt, QObject, QMetaObject, Slot, QUrl, QByteArray)
from qtpy.QtGui import QIcon, QKeySequence, QPixmap
from qtpy.QtWidgets import (QAction, QApplication, QFileDialog, QMainWindow,  QShortcut, QDialog, QLabel,
        QMdiArea, QMessageBox, QTextEdit, QWidget, QStyle, QStyleFactory, QActionGroup, QButtonGroup)                
from qtpy.QtWidgets import QMenu, QDockWidget, QListWidget
from qtpy import QtWidgets

from .. import config, gui

from .base import CheckMenu
from ..ezdock import overlay
from ..gcore.utils import getMenuAction, relax_menu_trace, relax_menu_text
from ..dialogs.main import NewPanelMenu, ShowMenu, WindowMenu
from ..dialogs.formlayout import fedit
from ..dialogs.about import AboutScreen

respath = Path(config['respath'])
logger = logging.getLogger(__name__)         
        
def debug_print(message):
    sys.__stdout__.write(f'{message}\n')
    sys.__stdout__.flush()        
    
class DragState(enum.Enum):
    placed = enum.auto()
    titlePressed = enum.auto()
    dragging = enum.auto()
    
        
class MainWindow(QMainWindow):
    """
    The Main QT Window.    
    """
    
    moveQueued = QtCore.Signal(object)
    
    def __init__(self, qapp, name, parentWinName=None):
        super(MainWindow, self).__init__()
        
        self.qapp = qapp
        self.name = name
        
        self.panels = qapp.panels              
        self.dragState = DragState.placed
        self.qapp.panels.ezm
        
        if not parentWinName is None:        
            self.setParent(self.qapp.windows[parentWinName])
            self.setWindowFlags(self.windowFlags() | Qt.Tool)   
        
        self.container = self.qapp.panels.ezm.new_container(self, self.name)    
        self.setCentralWidget(self.container)
                
        self.setWindowTitle(f'[{self.name}]')
        
        self.createMenus()
        self.createStatusBar()       
        
        #Hiding panelsDialog will disable all shortcuts
        sc = QShortcut(QKeySequence("Ctrl+Shift+Alt+F12"), self, 
            lambda: self.qapp.panelsDialog.setVisible(not self.qapp.panelsDialog.isVisible()))
        sc.setContext(Qt.ApplicationShortcut)

        self.moveQueued.connect(self.moveWindow, Qt.QueuedConnection)                
        
        self.activeCategory = None
        self.activePanId = None
        self.priorHoverButton = None
        
    def moveWindow(self, pos):
        self.move(pos)
        
    def setPanelInfo(self, panel=None):
        if panel is None:
            self.setWindowTitle(f'[{self.name}]')
            self.panelName.setText('')        
            self.activeCategory = None
            self.activePanId = None
            return
            
        self.setWindowTitle(f'[{self.name}] {panel.long_title}')
        self.panelName.setText(panel.short_title)        
        self.activeCategory = panel.category
        self.activePanId = panel.panid
                        
    def keyPressEvent(self, event):
        pass        
                        
    def createMenus(self):      
        self.layoutMenu = QtWidgets.QMenu()        
        
        panDiaAct = QAction("Panels Dialog", self, triggered=self.qapp.showDialog)
        panDiaAct.setIcon(QIcon(str(respath / 'icons' / 'px16' / 'application_view_gallery.png')))
        self.layoutMenu.addAction(panDiaAct)
        self.newPanelMenu = NewPanelMenu(self, showIcon=True)
        self.layoutMenu.addMenu(self.newPanelMenu)  
        
        self.panelMenu = ShowMenu(self, showIcon=True)
        self.layoutMenu.addMenu(self.panelMenu)       
        self.layoutMenu.addMenu(WindowMenu(self, showIcon=True))
        
        def addWindowMenuItem(caption, function, icon=None):
            keySequence = self.qapp.menuCallShortCuts.get('window', {}).get((relax_menu_text(caption),), None)
            if not keySequence is None:
                caption = f'{caption}\t{keySequence}'
            if isinstance(icon, str):
                icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / icon))
            if not icon is None:
                action = QAction(caption, self.windowMenu, triggered=function, icon=icon)
            else:
                action = QAction(caption, self.windowMenu, triggered=function)
            self.windowMenu.addAction(action)
        
        self.windowMenu = CheckMenu("&Layout Edit")
        self.windowMenu.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'layout_edit.png')))
        self.layoutMenu.addMenu(self.windowMenu)                   
            
        self.toolWinAction = QAction(self.winActionLabel(), self, triggered=self.asToolWindow)
        self.windowMenu.addAction(self.toolWinAction, checkcall=self.isToolWindow)
            
        addWindowMenuItem("Distribute", self.container.distribute, 'layouts_six_grid.png')        
        addWindowMenuItem("Drop In", lambda: self.qapp.panels.ezm.drop_in(self.container),
            'layouts_body_select.png')
        addWindowMenuItem("Screenshot to Clipboard", self.screenShot, 'lcd_tv_image.png')
        addWindowMenuItem("Cycle Tag Level", self.cycle_tag_level)
        addWindowMenuItem("Full Screen", self.fullScreen, 'view_fullscreen_view.png')
        addWindowMenuItem("Hide/Show Menu && Statusbar", self.toggleMenuStatusbar)                  
        addWindowMenuItem('Save Layout...', self.qapp.panelsDialog.layoutMenu.saveLayout)
        
        self.layoutMenu.addSeparator()
        self.qapp.panelsDialog.layoutMenu.addLayoutActions(self.layoutMenu)
        self.layoutMenu.addSeparator()

        act = QAction("E&xit", self, shortcut=QKeySequence.Quit,
            statusTip="Exit the application",
            triggered=QApplication.instance().quit)
        act.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'door_out.png')))
        self.layoutMenu.addAction(act)        
        
        self.panelsDialogBtn = QtWidgets.QToolButton(self)
        self.panelsDialogBtn.setIcon(QIcon(str(respath / 'icons' / 'px16' / 'application_view_gallery.png')))
        self.panelsDialogBtn.clicked.connect(self.qapp.showDialog)   
        self.panelsDialogBtn.setMenu(self.layoutMenu)
        self.panelsDialogBtn.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup) 
        self.menuBar().setCornerWidget(self.panelsDialogBtn, corner= Qt.TopLeftCorner)
        
        self.cycleTabBtn = QtWidgets.QToolButton(self)
        self.cycleTabBtn.setIcon(QIcon(str(respath / 'icons' / 'px16' / 'layout_content.png')))
        self.cycleTabBtn.clicked.connect(self.cycle_tag_level)
        self.cycleTabBtn.setMenu(self.windowMenu)
        self.cycleTabBtn.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)        
        self.menuBar().setCornerWidget(self.cycleTabBtn, corner= Qt.TopRightCorner)
        
    @property
    def windowName(self):
        return self.name
        
    def newPanelMenuPopup(self):
        self.newPanelMenu.exec_(QtGui.QCursor().pos())
        
    def cycle_tag_level(self):
        self.container.cycle_tag_level()
        
    def remove_panel_menu(self):
        main_childs = []
        for main_child in self.menuBar().children():
            main_childs.append(main_child)        
    
        self.menuBar().clear()

        for child in main_childs:
            if isinstance(child, QMenu):         
                self.menuBar().addMenu(child)  

    def screenShot(self):      
        pixmap = self.grab()
        qimage = pixmap.toImage()
        clipboard =  self.qapp.clipboard()
        clipboard.setImage(qimage)        
        
    def asToolWindow(self, windowName=None):
    
        if (not self.isToolWindow()) and windowName is None:
            winNames = [winname for winname, window in gui.qapp.windows.items()
                if not (window.isToolWindow() or winname == self.name)]
            winNames.append('None')
            winNameIndex = fedit([('Window', [1]+winNames)])[0]            
            winName = winNames[winNameIndex-1]            
        else:
            winName = windowName
            
        assert winName != self.name
        
        if winName in [None, 'None']:            
        
            self.setWindowFlags(self.windowFlags() & ~Qt.Tool)
            self.setParent(None)
            #setParent doc: Sets the parent of the widget to parent , and resets the window flags.
            self.toolWinAction.setText(self.winActionLabel())            
            
            #BUG
            # It seems that the window don't accept panel drops anymore
            # Only windows which had parent=None at __init__, seems to accept panel drops
            # Or windows with parent != None
            # But not windows are not reparent to None          
            # Removing the Tool flag before setting parent to None seems to solve it
            
            self.show()
            return
    
        self.setParent(self.qapp.windows[winName])
        self.toolWinAction.setText(self.winActionLabel())        
        self.setWindowFlags(self.windowFlags() | Qt.Tool)
        
        #It seems that the child behavior (on top of parent, minimize together with parent)
        #is lost as soon the parent is hidden and shown again
        #Toggling the stay on top flags restores the parent-child behavior
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        self.qapp.processEvents()
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowStaysOnTopHint)
        
        self.show()
        
    def winActionLabel(self):    
        if self.parent() is None:
            return 'Tool of...'
        else:
            return f'Tool of {self.parent().name}'        
        
    def isToolWindow(self):
        return (self.windowFlags() & Qt.Tool) == Qt.Tool        
            
    def createStatusBar(self):        
        self.panelName = QLabel('')
        self.panelName.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.panelName.customContextMenuRequested.connect(self.showPanelMenu)                
        self.statusBar().addWidget(self.panelName,1)      
        
    def showPanelMenu(self, point=None):
        self.panelMenu.exec_(QtGui.QCursor().pos())
                                               
    def fullScreen(self):
        if not self.isFullScreen():
            self.showFullScreen()
        else:
            self.showNormal() 

    def toggleMenuStatusbar(self):
        if self.menuBar().isVisible():
            self.menuBar().hide()
            self.statusBar().hide()
        else:
            self.menuBar().show()
            self.statusBar().show()
        
    def unregister(self):        
        #Detach all tool windows of this window
        for window in self.qapp.windows.values():
                if window.parent() == self:
                    window.asToolWindow(None)
                                    
        #Unregister this window
        self.qapp.windows.pop(self.name)
        self.qapp.panels.ezm.containers.pop(self.name)
        self.deleteLater()        
        self.qapp.hideWindow()

    def getWidgetUnderMouse(self):
        '''
        Get the widget outside this window under the current mouse position.        
        '''
        self.setAttribute(Qt.WA_TransparentForMouseEvents, True)        
        pos = QtGui.QCursor.pos()
        widget = gui.qapp.widgetAt(pos)
        self.setAttribute(Qt.WA_TransparentForMouseEvents, False)
        if widget is None: return None, None
        locPos = widget.mapFromGlobal(pos)
        return widget, locPos
        
    def moveEvent(self, event):
        if self.dragState == DragState.placed: return
        
        elif self.dragState == DragState.titlePressed:
            self.dragState = DragState.dragging
            self.setWindowOpacity(0.5)
            self.qapp.panels.ezm.drop_in(self.container, hide=False, allWindows=self.dragToAllWindows)            
            
        elif self.dragState == DragState.dragging:                                
            widget, locPos = self.getWidgetUnderMouse()        
            
            if widget is None: return
            
            if self.priorHoverButton == widget: return
            
            if not self.priorHoverButton is None:
                self.priorHoverButton.endPreview()
                # It seems that sometimes, the priorHoverButton is not valid?
                # Traceback (most recent call last):
                 # File "c:\tools\gh2\venv\lib\site-packages\ghawk2\panels\window.py", line 320, in moveEvent
                 # self.priorHoverButton.endPreview()
                 # File "c:\tools\gh2\venv\lib\site-packages\ghawk2\ezdock\overlay.py", line 59, in endPreview
                 # self.parent().endPreview() 
                # RuntimeError: Internal C++ object (HoverButton) already deleted.
                self.priorHoverButton = None                
            
            if isinstance(widget, overlay.HoverButton):
                widget.startPreview()
                self.priorHoverButton = widget            
    
    def event(self, event):                       
        if self.dragState == DragState.placed:
            if event.type() == QtCore.QEvent.NonClientAreaMouseButtonPress:
                if self.isToolWindow():
                    if event.modifiers() & QtCore.Qt.ControlModifier:
                        self.startMoving(True)
                    else:
                        self.startMoving(False)
                elif event.modifiers() & QtCore.Qt.ControlModifier:
                    self.startMoving(True)
                else:
                    return super().event(event)
                return True
            
        elif event.type() == QtCore.QEvent.NonClientAreaMouseButtonDblClick \
            or event.type() == QtCore.QEvent.Resize:
            self.windowPlaced(canceled=True)            
            
        elif event.type() == QtCore.QEvent.NonClientAreaMouseButtonRelease:            
            self.windowPlaced() 
            return True
            
        return super().event(event)
        
    def startMoving(self, allWindows=True):
        self.dragState = DragState.titlePressed          
        self.dragToAllWindows = allWindows 
        self.priorHoverButton = None
        
    def windowPlaced(self, canceled=False):            
        if not canceled:
            widget, locPos = self.getWidgetUnderMouse()
            
            if isinstance(widget, overlay.HoverButton):
                widget.clicked.emit()
                self.qapp.deleteEmptyWindows()
            else:
                canceled = True
                
        if canceled:
            self.qapp.panels.ezm.hide_overlays()                 
                
        self.setWindowOpacity(1.0)
        self.dragState = DragState.placed         
        self.priorHoverButton = None
        
    def closeEvent(self, event):
        self.qapp.hideWindow(self)
        event.accept()


from ...core.gui_proxy import GuiProxyBase, StaticGuiCall, gui
        
class CmdHistGuiProxy(GuiProxyBase):    
    category = 'cmdhist'
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.cmdhist = self
    



import math
import numpy as np

from qtpy import QtCore, QtGui, QtWidgets

from ...graphics.plotview import PlotView
from ...graphics.rulers import TickedRuler, Axis
from ...graphics.functions import arrayToQPath

from ...utils.ticks import tickValues

class ProfileGraphicView(PlotView):

    def __init__(self, parent=None):
        super().__init__(parent)                                    
           
                   
class ProfilerPanel(QtWidgets.QWidget):

    def __init__(self, parent, direction, imviewer):
        super().__init__(parent=parent)
        
        self.imagePanel = imviewer
        self.scene = QtWidgets.QGraphicsScene()
        self.view = ProfileGraphicView(self)
        self.view.setScene(self.scene)            
        
        vbox = QtWidgets.QVBoxLayout(self)
        vbox.setContentsMargins(0,0,0,0)
        vbox.addWidget(self.view)        
        self.setLayout(vbox)        
       
        if direction == 'x':
            self.direction = 0
            self.view.fixScaleX = True
            self.view.scale[1] = -self.view.scale[1]
            self.view.updateMatrix()
            self.setMinimumHeight(20)
        else:
            self.direction = 90
            self.view.fixScaleY = True
            self.view.scale[0] = -self.view.scale[0]
            self.view.updateMatrix()
            self.setMinimumWidth(20)
                                                       
        #self.view.matrixUpdated.connect(self.recalcSlices)
        self.view.matrixUpdated.connect(self.redrawSlices)
        self.view.matrixUpdated.connect(self.repositionRuler)        
        self.view.matrixUpdated.connect(self.repositionYAxis)
        self.view.doubleClicked.connect(self.zoomAuto)           
                
        self.meanProfileCurve = None
        self.roiProfileCurve = None
        self.pixProfileCurve = None

        
    def zoomAuto(self):
            
        top = bottom = left = right = None
        for curve in [self.pixProfileCurve, self.roiProfileCurve, self.meanProfileCurve]:
            if curve is None:
                continue
            r = curve.boundingRect()
            top = r.top() if top is None else min(r.top(), top)
            bottom = r.bottom() if bottom is None else max(r.bottom(), bottom)
            left = r.left() if left is None else min(r.left(), left)
            right = r.right()if right is None else max(r.right(), right)
            
        if top is None:
            return           
            
        height = bottom - top
        width = right - left
            
        if self.direction == 0:                          
            if height == 0:                
                self.view.setYCenter(top)
            else:
                scale = -(self.height() - 20) / max(height, 1e-9)
                self.view.setYPosScale(bottom, scale)
        elif self.direction == 90: 
            if width == 0:
                self.view.setXCenter(left)
            else:
                scale = -(self.width() - 20)/ max(width, 1e-9)
                self.view.setXPosScale(right, scale)
    
    def refresh(self):
        self.removeAllItems()

        self.createRuler()
        self.createYAxis()
        self.createGrid()                                
        
        if self.meanProfileCurve is not None:
            self.scene.addItem(self.meanProfileCurve)
        if self.pixProfileCurve is not None:
            self.scene.addItem(self.pixProfileCurve)
                    
        self.view.refresh()       
        self.redrawSlices()
        self.repositionRuler()
        self.repositionYAxis()                
        
        
    def removeAllItems(self):
        for i in self.scene.items():
            #only need to remove top level items
            if i.parentItem() == None:
                self.scene.removeItem(i)
        
    def createYAxis(self):
        if not self.yAxisEnabled:
            self.yAxis = None
            return 0            
            
        if self.direction == 0:        
            scaleY = -self.view.scale[1]
            sceneTopLeft = self.view.mapToScene(0,0)
            sceneBotRight = self.view.mapToScene(self.width()-1, self.height()-1)            
            self.stopY = sceneTopLeft.y()
            self.startY = sceneBotRight.y()                        
            self.thicksY = tickValues(self.startY, self.stopY, scaleY)           
            self.yAxis = Axis(0, self.startY, self.stopY, self.thicksY)
        elif self.direction == 90:
            scaleY = -self.view.scale[0]
            sceneTopLeft = self.view.mapToScene(0,0)
            sceneBotRight = self.view.mapToScene(self.width()-1, self.height()-1)            
            self.stopY = sceneTopLeft.x()
            self.startY = sceneBotRight.x()                        
            self.thicksY = tickValues(self.startY, self.stopY, scaleY)
            self.yAxis = Axis(90, self.startY, self.stopY, self.thicksY)
            
        #self.yAxis.setZValue(-1)
        self.yAxis.setZValue(0.9)
        self.scene.addItem(self.yAxis)
        
    def createRuler(self):
        #self.ruler = Ruler(self.direction, self.startX, self.stopX, self.stepX)
        
        if self.direction == 0:
            scaleX = self.view.scale[0]
            
            if self.imagePanel.dispOffsetX < 0:
                self.startX = 0
                self.stopX = self.imagePanel.dispOffsetX + self.width() / scaleX
            else:
                self.startX = self.imagePanel.dispOffsetX
                self.stopX = math.ceil(self.startX + self.width() / scaleX)
                
            self.stopX = min(self.imagePanel.vd.width, self.stopX)                                                                      
            self.ruler = TickedRuler(0, self.startX, self.stopX, scaleX, noDecimals=True)        
            
        elif self.direction == 90:
            scaleX = self.view.scale[1]
            
            if self.imagePanel.dispOffsetY < 0:
                self.startX = 0
                self.stopX = self.imagePanel.dispOffsetY + self.height() / scaleX
            else:
                self.startX = self.imagePanel.dispOffsetY
                self.stopX = math.ceil(self.startX + self.height() / scaleX)
                
            self.stopX = min(self.imagePanel.vd.height, self.stopX)            
            self.ruler = TickedRuler(90, self.startX, self.stopX, scaleX, noDecimals=True)        
            
        self.ruler.setZValue(1)
        self.scene.addItem(self.ruler)
        
    def repositionRuler(self):
        tr = self.view.mapToScene(self.width()-20, self.height()-20)
        if self.direction == 0:
            self.ruler.setY(tr.y())
        elif self.direction == 90:
            self.ruler.setX(tr.x())
                  
    def repositionYAxis(self):
        if self.yAxis == None:
            return 0            
        tr = self.view.mapToScene(50, 10)
        if self.direction == 0:
            self.yAxis.setX(tr.x())
        elif self.direction == 90:
            self.yAxis.setY(tr.y())

    def zoomToImage(self):
        scale = self.imagePanel.zoomValue
        
        if self.direction == 0:
            x = self.imagePanel.dispOffsetX        
            self.view.setXPosScale(x, scale)
        elif self.direction == 90:
            y = self.imagePanel.dispOffsetY       
            self.view.setYPosScale(y, scale)

        
    def addPlot(self, x, y, color=None, z=0):
        curve = self.createCurve(x, y, color, z)
        self.curves.append(curve)
        
    def drawMeanProfile(self, x, y):
        if self.meanProfileCurve is not None:
            self.scene.removeItem(self.meanProfileCurve)
        
        self.meanProfileCurve = self.createCurve(x, y, color=QtCore.Qt.blue, z=0.5)
        self.scene.addItem(self.meanProfileCurve)               
        
    def drawRoiProfile(self, x, y):
        if self.roiProfileCurve is not None:
            self.scene.removeItem(self.roiProfileCurve)
        
        self.roiProfileCurve = self.createCurve(x, y, color=QtCore.Qt.darkGreen, z=0.25)
        self.scene.addItem(self.roiProfileCurve)          
        
    def drawPixelProfile(self, x, y):
        self.removePixelProfile()        
        self.pixProfileCurve = self.createCurve(x, y, color=QtCore.Qt.red, z=0)
        self.scene.addItem(self.pixProfileCurve)                

    def removeRoiProfile(self):
        if self.roiProfileCurve is not None:
            self.scene.removeItem(self.roiProfileCurve)
            self.roiProfileCurve = None
            
    def removePixelProfile(self):
        if self.pixProfileCurve is not None:
            self.scene.removeItem(self.pixProfileCurve)
            self.pixProfileCurve = None            
            
    def removeMeanProfile(self):
        if self.meanProfileCurve is not None:
            self.scene.removeItem(self.meanProfileCurve)
            self.meanProfileCurve = None            
            
        
    def createCurve(self, x, y, color=None, z=0):
        if color == None:
            pen = QtGui.QPen(QtCore.Qt.blue, 0, QtCore.Qt.SolidLine)
        else:
            pen = QtGui.QPen(color, 0, QtCore.Qt.SolidLine)
        #first create a Path        
                
        if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
            if self.direction == 0:
                path = arrayToQPath(x, y)
            elif self.direction == 90:
                path = arrayToQPath(y, x)
        else:
            path = QtGui.QPainterPath()
        
            if self.direction == 0:
                path.moveTo(x[0], y[0])
                for i in range(1, len(y)):
                    path.lineTo(x[i], y[i])     
            elif self.direction == 90:
                path.moveTo(y[0], x[0])
                for i in range(1, len(y)):
                    path.lineTo(y[i], x[i])
                
        #transform the Path to a PathItem                                    
        curve = QtWidgets.QGraphicsPathItem(path)
        if z != 0:
            curve.setZValue(z)
        curve.setPen(pen)
        
        return curve
                        
    def clear(self):
        self.meanProfileCurve = None
        self.roiProfileCurve = None
        self.pixProfileCurve = None
        
    def showOnlyRuler(self):
        if self.direction == 0:
            self.setFixedHeight(20)
        elif self.direction == 90:
            self.setFixedWidth(20)          
        self.gridEnabled = False
        self.yAxisEnabled = False
        self.plotsEnabled = False
        self.refresh()
        
    def showAll(self):
        self.gridEnabled = True
        self.yAxisEnabled = True
        self.plotsEnabled = True
        self.refresh()
        
    def removelast(self):
        if len(self.curves) > 0:
            lastCurve = self.curves[-1]
            if lastCurve in self.scene.items():
                self.scene.removeItem(lastCurve)
            self.curves.pop()
        
    def createGrid(self):
        self.grid = []
        if not self.gridEnabled:
            return 0
            
        pens = []
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))        
        paths = []
                
                           
        for thicklevel in range(3):
            paths.append(QtGui.QPainterPath())
            for i in self.ruler.thicks[thicklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(i, self.startY)
                    paths[-1].lineTo(i, self.stopY)               
                else:
                    paths[-1].moveTo(self.startY, i)
                    paths[-1].lineTo(self.stopY, i) 
                               
        for thicklevel in range(3):        
            paths.append(QtGui.QPainterPath())  
            for i in self.thicksY[thicklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(self.startX, i)
                    paths[-1].lineTo(self.stopX, i)
                else:
                    paths[-1].moveTo(i, self.startX)
                    paths[-1].lineTo(i, self.stopX)
                
        for i in range(len(paths)):
            self.grid.append(QtWidgets.QGraphicsPathItem(paths[i]))
            self.grid[-1].setPen(pens[i])
            self.grid[-1].setZValue(-2)
            self.scene.addItem(self.grid[-1])                                      


        
    def redrawSlices(self):        
        self.scene.removeItem(self.ruler)
        self.createRuler()        
        
        if self.yAxis != None:
            self.scene.removeItem(self.yAxis)
        self.createYAxis()
        
        for items in self.grid:
            self.scene.removeItem(items)            
        self.createGrid()   

        

        



from qtpy import QtCore, QtGui, QtOpenGL, QtWidgets
from .point import Point
from . import functions as fn

QtSignal = QtCore.Signal

HAVE_OPENGL = True

class PlotView(QtWidgets.QGraphicsView):

    """
    Re-implementation of QGraphicsView without scrollbars.

    Allow unambiguous control of the viewed coordinate range.
    """

    matrixUpdated = QtSignal()
    doubleClicked = QtSignal()
    
    scaleXUpdated = QtSignal()
    panXUpdated = QtSignal()
    scaleYUpdated = QtSignal()
    panYUpdated = QtSignal()

    def __init__(self, parent=None, background='default'):
        super().__init__(parent)

        # There is the experimental options of using OpenGl
        self.useOpenGL(False)
        self.setCacheMode(self.CacheBackground)
        self.setBackground(background)
        
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setResizeAnchor(QtWidgets.QGraphicsView.NoAnchor)
        #self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)

        self.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
        
        #self.setViewportUpdateMode(QtWidgets.QGraphicsView.MinimalViewportUpdate)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.NoViewportUpdate)
        
        #self.setStyleSheet( "QGraphicsView { border-style: none; }" )
        
        #self.setMouseTracking(True)
        
        #self.range = QtCore.QRectF(0, 0, 1, 1)
        self.scale = [100 / 2**16, 100 / 2**16]
        self.center = [2**15, 2**15]
        self.lastMousePos = None
        
        self.fixScaleX = False
        self.fixScaleY = False
        
        fullrange = QtCore.QRectF(-1e6, -1e6, 2e6, 2e6)
        self.setSceneRect(fullrange)
        self.updateMatrix()
        
        self.initMenu()
        
    @property
    def auto_zoom(self):
        """
        Check whether the 'Auto Zoom' checkbox is checked.

        :return bool: True when auto zoom is checked / enabled.
        """
        return self.autoAction.isChecked()
        
    def initMenu(self):  
        self.menu = QtWidgets.QMenu(self)
        
        self.autoAction = QtWidgets.QAction('Auto Zoom', self)
        self.autoAction.setCheckable(True)
        self.autoAction.setChecked(True)
        self.autoAction.triggered.connect(self.toggleAutoZoom)
        self.menu.addAction(self.autoAction)       
        
    def toggleAutoZoom(self):
        #just toggle the check boxs
        pass     
        
    def useOpenGL(self, b=True):
        if b:
            if not HAVE_OPENGL:
                raise Exception("Requested to use OpenGL with QGraphicsView, but QtOpenGL module is not available.")
            v = QtOpenGL.QGLWidget()
        else:
            v = QtWidgets.QWidget()
            
        self.setViewport(v)

    def setBackground(self, background):
        """
        Set the background color of the GraphicsView.
        To make the background transparent, use background=None.
        """
        self._background = background
        if background == 'default':
            background = (250, 250, 250)
        if background is None:
            self.setBackgroundRole(QtGui.QPalette.NoRole)
        else:
            brush = fn.mkBrush(background)
            self.setBackgroundBrush(brush)
        
    def updateMatrix(self, propagate=True):        
        #t = self.transform()
        self.limitRefreshRange()        
        self.setTransform(QtGui.QTransform(\
            self.scale[0], 0  , 0,\
            0  , self.scale[1], 0,\
            0  , 0  , 1))
        
        #self.ensureVisible(self.range,0,0)        
        self.centerOn(*self.center)       
        #self.setSceneRect(self.range)
        #self.fitInView(self.range, QtCore.Qt.IgnoreAspectRatio)                        
        self.matrixUpdated.emit()
        self.viewport().update()
        #self.updateSceneRect(self.range)
        
    def limitRefreshRange(self):
        self.range = QtCore.QRectF()
        self.range.setWidth((self.width() + 20)/ self.scale[0])
        self.range.setHeight((self.height() + 20) / self.scale[1])
        self.range.moveCenter(QtCore.QPointF(*self.center))        
        self.setSceneRect(self.range)    
               
    def translate(self, dx, dy):
        self.center = [self.center[0] +dx, self.center[1] + dy]
        self.updateMatrix()
        if dx != 0:
            self.panXUpdated.emit()
        if dy != 0:
            self.panYUpdated.emit()
            
    def setXCenter(self, pos):
        self.center[0] = pos
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            self.scale[0], 0      , 0,\
            0    , t.m22(), 0,\
            0    , 0      , 1))                    
        self.centerOn(*self.center)
        
        self.scaleXUpdated.emit()
        self.panXUpdated.emit()
        self.matrixUpdated.emit()
        self.viewport().update()            
        
                
    def setXPosScale(self, pos, scale):
        if scale == 0:
            return
        self.scale[0] = scale
        self.center[0] = pos + self.width() / 2.0 / scale 
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            scale, 0      , 0,\
            0    , t.m22(), 0,\
            0    , 0      , 1))                    
        self.centerOn(*self.center)
        
        self.scaleXUpdated.emit()
        self.panXUpdated.emit()
        self.matrixUpdated.emit()
        self.viewport().update()
        #self.updateSceneRect(self.range)
        
    def setYCenter(self, pos):
        self.center[1] = pos 
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            t.m11(), 0    , 0,\
            0      , self.scale[1], 0,\
            0      , 0    , 1))                    
        self.centerOn(*self.center)
        
        self.scaleYUpdated.emit()
        self.panYUpdated.emit()
        self.matrixUpdated.emit()   
        self.viewport().update()        
        
        
    def setYPosScale(self, pos, scale):
        if scale == 0:
            return
        self.scale[1] = scale
        self.center[1] = pos + self.height() / 2.0 / scale 
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            t.m11(), 0    , 0,\
            0      , scale, 0,\
            0      , 0    , 1))                    
        self.centerOn(*self.center)
        
        self.scaleYUpdated.emit()
        self.panYUpdated.emit()
        self.matrixUpdated.emit()   
        self.viewport().update()
        #self.updateSceneRect(self.range)        
        
    def refresh(self):
        self.viewport().update()
        
    def setScale(self, sx, sy):
        self.scale = [sx, sy]        
        self.updateMatrix()        

    def mousePressEvent(self, ev):
        #QtWidgets.QGraphicsView.mousePressEvent(self, ev)
        
        if (ev.buttons() == QtCore.Qt.LeftButton):
            self.lastMousePos = Point(ev.pos())
            self.mousePressPos = ev.pos()

        if (ev.buttons() == QtCore.Qt.RightButton):
            pos = QtGui.QCursor.pos()
            self.menu.exec_(pos)  
            
        self.clickAccepted = ev.isAccepted()                  
               
    def mouseMoveEvent(self, ev):
    
        if self.lastMousePos is None:
            self.lastMousePos = Point(ev.pos())
            
        delta = Point(ev.pos() - self.lastMousePos.toPoint())
        self.lastMousePos = Point(ev.pos())

        # QtWidgets.QGraphicsView.mouseMoveEvent(self, ev)
        
        if ev.buttons() in [QtCore.Qt.MidButton, QtCore.Qt.LeftButton]:  ## Allow panning by left or mid button.
            px = self.pixelSize()
            tr = -delta * px
            if self.fixScaleX:
                tr[0] = 0
            if self.fixScaleY:
                tr[1] = 0
            self.translate(tr[0], tr[1])
                
    def mouseDoubleClickEvent(self, event):
        self.doubleClicked.emit()
            
    def wheelEvent(self, ev):
        # QtWidgets.QGraphicsView.wheelEvent(self, ev)
        if not self.fixScaleX and not self.fixScaleY:
            self.scale[0] = self.scale[0] * 1.001 ** ev.delta()
            self.scale[1] = self.scale[1] * 1.001 ** ev.delta()
            self.updateMatrix()
            self.scaleXUpdated.emit()
            self.panXUpdated.emit()
            self.scaleYUpdated.emit()
            self.panYUpdated.emit()
        elif self.fixScaleX and not self.fixScaleY:                
                self.scale[1] = self.scale[1] * 1.001 ** ev.delta()
                self.updateMatrix()
                self.scaleXUpdated.emit()
                self.panXUpdated.emit()
        elif not self.fixScaleX and self.fixScaleY:                
                self.scale[0] = self.scale[0] * 1.001 ** ev.delta()        
                self.updateMatrix()
                self.scaleXUpdated.emit()
                self.panYUpdated.emit()

    def pixelSize(self):
        """Return vector with the length and width of one view pixel in scene coordinates"""
        p0 = Point(0,0)
        p1 = Point(1,1)
        tr = self.transform().inverted()[0]
        p01 = tr.map(p0)
        p11 = tr.map(p1)
        return Point(p11 - p01)



import numpy as np
import numba

@numba.njit(cache=True)
def bincount2d(array, minlength=65536):          
    hist = np.bincount(array.ravel(), minlength=minlength)        
    return hist

@numba.njit(parallel=True, cache=True)    
def map_values_mono(source, target, mapvector):
    length = len(source)
    for i in numba.prange(length):
        value = source[i]
        target[i] = mapvector[value]
        
@numba.njit(parallel=True, cache=True)    
def map_values_rgb(source, target, mapvector):
    height, width, channels = source.shape
    for i in numba.prange(height):
        for j in numba.prange(width):
            target[i,j,0] = mapvector[source[i,j,0]]        
            target[i,j,1] = mapvector[source[i,j,1]]        
            target[i,j,2] = mapvector[source[i,j,2]]        
        
@numba.njit(parallel=True, cache=True)    
def map_values_rgbswap(source, target, mapvector):
    height, width, channels = source.shape
    for i in numba.prange(height):
        for j in numba.prange(width):
            target[i,j,2] = mapvector[source[i,j,0]]        
            target[i,j,1] = mapvector[source[i,j,1]]        
            target[i,j,0] = mapvector[source[i,j,2]]        
            target[i,j,3] = 255
            
@numba.njit(parallel=True, cache=True)    
def get_min_max(source):
    return source.min(), source.max()

def nb_float_offset_gain_gamma_8bit(array, offset=0, gain=1, gamma=1):
    procarr = np.ndarray(array.shape, 'uint8')
    nb_float_offset_gain_gamma_loop(array, procarr, offset, gain, gamma * 1.0)
    return procarr
    
@numba.njit(parallel=True, cache=True)     
def nb_float_offset_gain_gamma_loop(source, target, offset=0, gain=1, gamma=1):        
    height, width = source.shape
    gscale = 255 ** (1 - gamma)
    
    for i in numba.prange(height):
        for j in numba.prange(width):
            tmp = (source[i,j] - offset) * (gain * 256)
            if gamma == 1.0:
                if tmp < 0:
                    target[i,j] = 0
                elif tmp > 255:
                    target[i,j] = 255
                else:
                    target[i,j] = tmp
            else:
                tmp2 = tmp ** gamma * gscale
                if tmp2 < 0:
                    target[i,j] = 0
                elif tmp2 > 255:
                    target[i,j] = 255
                else:
                    target[i,j] = tmp2


from qtpy import QtCore, QtGui, QtWidgets

class FileDialog(QtWidgets.QDialog):

    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        
        self.treeView = QtWidgets.QTreeView(self)
        hbox.addWidget(self.treeView)       
        
        #self.treeView.setGeometry(0,0, 600, 800)
        self.dirModel = QtWidgets.QFileSystemModel()
        self.path = "C:\\"
        self.dirModel.setRootPath(self.path)
        self.dirModel.setFilter(QtCore.QDir.AllDirs | QtCore.QDir.Dirs | QtCore.QDir.NoDotAndDotDot | QtCore.QDir.Hidden)
        self.treeView.setModel(self.dirModel)
        #self.treeView.setRootIndex(self.dirModel.index(self.path))
        #self.treeView.expandAll()
        self.treeView.clicked.connect(self.on_treeView_clicked)

        #self.listView = QtWidgets.QListView(self)
        #self.listView = QtWidgets.QTreeView(self)
        self.listView = QtWidgets.QTableView(self)
        hbox.addWidget(self.listView)
        
        self.fileModel = QtWidgets.QFileSystemModel()
        #self.fileModel.setFilter(QtCore.QDir.Files | QtCore.QDir.Dirs | QtCore.QDir.Hidden)
        self.fileModel.setFilter(QtCore.QDir.Files | QtCore.QDir.Hidden)
        self.listView.setModel(self.fileModel)

        #self.setGeometry(300, 300, 600, 800)
        self.setWindowTitle('Select File')        
        
    def on_treeView_clicked(self, index):
        # // TreeView clicked
        # // 1. We need to extract path
        # // 2. Set that path into our ListView

        # Get the full path of the item that's user clicked on
        path = self.dirModel.fileInfo(index).absoluteFilePath()
        index = self.fileModel.setRootPath(path)
        self.listView.setRootIndex(index)
        
    def ok(self):
        self.updatePaths()
        self.done(QtWidgets.QDialog.DialogCode.Accepted)
        
    def cancel(self):
        self.done(QtWidgets.QDialog.DialogCode.Rejected)        
        
    


import sys
import platform
import code
from pathlib import Path

def interact(workspace=None, banner=None, exitmsg=None, completer='standard'):    
    if workspace is None:
        frame = sys._getframe(1)
        globs = frame.f_globals
        locs = frame.f_locals
        workspace = globs
        workspace.update(locs) 

    try:
        import readline
        if completer=='standard':
            from rlcompleter import Completer
            readline.set_completer(Completer(workspace).complete)          
        elif completer=='key':
            from .completer import Completer
            readline.set_completer_delims(' \t\n\\`@$><=;|&{(')
            readline.set_completer(Completer(workspace).complete)
        else:
            raise AttributeError(f'Invalid completer {completer}')
                          
        readline.parse_and_bind('tab:complete')
    except:
        print('Could not start auto complete!')
                                            
    shell = code.InteractiveConsole(workspace)
    shell.interact(banner=banner, exitmsg=exitmsg)          


import os
import time
import collections
from pathlib import Path
import types
from collections.abc import Iterable
import queue
import logging

import numpy as np

logger = logging.getLogger(__name__)

try:
    import scipy
    import scipy.ndimage
    has_scipy = True

except:
    has_scipy = False

try:
    import imageio
    has_imafio= True

except:
    has_imafio = False

from ... import config, gui

if has_imafio:
    if not config.get("path_imageio_freeimage_lib", None) is None:
        if os.getenv("IMAGEIO_FREEIMAGE_LIB", None) is None:
            os.environ["IMAGEIO_FREEIMAGE_LIB"] = config.get("path_imageio_freeimage_lib")

    try:
        imageio.plugins._freeimage.get_freeimage_lib()

    except Exception as ex:
        logger.warning('Could not load freeimage dll')
        logger.warning(str(ex))
        logger.warning('Automatic download can be a problem when using VPN')
        logger.warning("Download the dll's from https://github.com/imageio/imageio-binaries/tree/master/freeimage/")
        logger.warning(f'And place it in {imageio.core.appdata_dir("imageio")}/freeimage')

        #You can also use a system environmental variable
        #IMAGEIO_FREEIMAGE_LIB=<the location>\FreeImage-3.18.0-win64.dll

    #The effective dll is refered at
    #imageio.plugins.freeimage.fi.lib

    #Prefer freeimage above pil
    #Freeimage seems to be a lot faster then pil
    imageio.formats.sort('-FI', '-PIL')

    FILTERS_NAMES = collections.OrderedDict()
    FILTERS_NAMES['All Formats (*)'] = None

    for fmt in imageio.formats._formats_sorted:
        filter = f'{fmt.name} - {fmt.description} (' + ' '.join(f'*{fmt}' for fmt in fmt.extensions) + ')'
        FILTERS_NAMES[filter] = fmt.name

    IMAFIO_QT_READ_FILTERS = ';;'.join(FILTERS_NAMES.keys())
    IMAFIO_QT_WRITE_FILTERS = ';;'.join(FILTERS_NAMES.keys())    
    IMAFIO_QT_WRITE_FILTER_DEFAULT = "TIFF-FI - Tagged Image File Format (*.tif *.tiff)"


from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Qt, Signal, QUrl
from qtpy.QtGui import QFont, QTextCursor, QPainter, QPixmap, QCursor, QPalette, QColor, QKeySequence
from qtpy.QtWidgets import (QApplication, QAction, QMainWindow, QPlainTextEdit, QSplitter, QVBoxLayout, QHBoxLayout,
    QMessageBox, QTextEdit, QLabel, QWidget, QStyle, QStyleFactory, QLineEdit, QShortcut, QMenu, QStatusBar, QColorDialog)

from ...panels import BasePanel, thisPanel, CheckMenu
from ...panels.base import MyStatusBar, selectThisPanel
from ...dialogs.formlayout import fedit
from ...dialogs.colormap import ColorMapDialog
from ...widgets.grid import GridSplitter
from ...utils import lazyf, clip_array
from ...utils import imconvert
from ...gcore.utils import ActionArguments

from .profile import ProfilerPanel
from .blueprint import make_thumbnail
from .demosaic import bayer_split
from .quantiles import get_sigma_range_for_hist
from .spectrogram import spectr_hori, spectr_vert
from .imgdata import ImageData
from .roi import SelRoiWidget


ZOOM_VALUES = [
     0.005, 0.0064 , 0.008,
     0.01 , 0.0125 , 0.016,
     0.02 , 0.025  , 0.032,
     0.04 , 0.05   , 0.064,
     0.08 , 0.10   , 0.125,
     0.16 , 0.20   , 0.250,
     0.32 , 0.40   , 0.5  ,
     0.64 , 0.80   , 1.0  ,
     1.25 , 1.60   , 2.0  ,
     2.50 , 3.20   , 4.0  ,
     5.00 , 6.40   , 8.0  ,
    10.00 , 12.5   , 16.0 ,
    20.00 , 25.0   , 32.0 ,
    40.00 , 50.0   , 64.0 ,
    80.00 ,100.0   ,125.0 ,
    160.0 ,200.0   ,250.0]

here = Path(__file__).parent.absolute()
respath = Path(config['respath'])
#sck = config['shortcuts']

class StatusPanel(MyStatusBar):

    offsetGainEdited = Signal(str, str, str)
    zoomEdited = Signal(float)

    def __init__(self, parent):
        super().__init__(parent=parent)

        self.panel = parent

        console_font = QFont('Consolas', pointSize=config['console']['fontsize'])

        self.chooseValFormat = QMenu('Value Format', self)
        self.chooseValFormat.addAction(QAction("Decimal", self, triggered=lambda: self.set_val_format('dec')))
        self.chooseValFormat.addAction(QAction("Hex", self, triggered=lambda: self.set_val_format('hex')))
        self.chooseValFormat.addAction(QAction("Binary", self, triggered=lambda: self.set_val_format('bin')))

        self.set_val_format('dec')

        self.zoomOutBtn = QtWidgets.QPushButton(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'bullet_toggle_minus.png')), None, self)
        self.zoom = QLineEdit('100')
        self.zoom.keyPressEvent = self.zoomKeyPressEvent
        self.zoomInBtn = QtWidgets.QPushButton(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'bullet_toggle_plus.png')), None, self)
        self.xy = QLabel('0,0')
        self.vallab = QLabel('val')
        self.val = QLineEdit('0')
        self.val.setFont(console_font)
        self.val.setReadOnly(True)
        self.val.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.val.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.val.customContextMenuRequested.connect(lambda: self.chooseValFormat.exec_(QtGui.QCursor().pos()))

        self.addWidget(self.zoomOutBtn)
        self.addWidget(self.zoom, 1)
        self.addWidget(self.zoomInBtn)
        self.addWidget(self.xy, 1)
        self.addWidget(self.vallab, 1, Qt.AlignRight)
        self.addWidget(self.val, 2)

        self.offsetlab = QLabel('offset')
        self.offset = QLineEdit('0')
        self.offset.keyPressEvent = types.MethodType(offsetGainKeyPressEvent, self.offset)
        self.offset.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.gainlab = QLabel('gain')
        self.gain = QLineEdit('1')
        self.gain.keyPressEvent = types.MethodType(offsetGainKeyPressEvent, self.gain)
        self.gain.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.gammalab = QLabel('gamma')
        self.gamma = QLineEdit('1')
        self.gamma.keyPressEvent = types.MethodType(offsetGainKeyPressEvent, self.gamma)
        self.gamma.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        self.addWidget(self.offsetlab, 1, Qt.AlignRight)
        self.addWidget(self.offset, 1)
        self.addWidget(self.gainlab, 1, Qt.AlignRight)
        self.addWidget(self.gain, 1)
        self.addWidget(self.gammalab, 1, Qt.AlignRight)
        self.addWidget(self.gamma, 1)

        self.zoomOutBtn.clicked.connect(parent.zoomOut)
        self.zoomInBtn.clicked.connect(parent.zoomIn)
        self.offsetGainEdited.connect(parent.changeOffsetGain)
        self.zoomEdited.connect(parent.setZoomValue)

    def set_val_format(self, fmt='dec'):
        self.val_format = fmt
        self.val_item_format = None

    def set_val_item_format(self):
        if self.val_format == 'dec':
            self.val_item_format = '{v}'
        elif self.val_format == 'hex':
            self.val_item_format = '0x{v:X}'
        elif self.val_format == 'bin':
            self.val_item_format = '{v:b}'

    def set_xy_val(self, x, y, val=None):
        self.xy.setText(f'xy:{x:d},{y:d} ')

        if self.val_item_format is None:
            self.set_val_item_format()

        if not val is None:
            try:
                if isinstance(val, Iterable):
                    text = '[' + ' '.join(lazyf(self.val_item_format) for v in val) + ']'
                    self.val.setText(text)
                else:
                    v = val
                    self.val.setText(lazyf(self.val_item_format))
            except:
                self.val.setText(str(val))

    def setOffsetGainInfo(self, offset, gain, gamma):
        self.offset.setText(f'{offset:8.6g}')
        self.gain.setText(f'{gain:8.6g}')
        self.gamma.setText(f'{gamma:8.6g}')

    def set_zoom(self, value):
        self.zoom.setText(f'{value*100:.2f}')

    def zoomKeyPressEvent(self, event):
        key_enter = (event.key() == Qt.Key_Return) or \
            (event.key() == Qt.Key_Enter)

        statpan = self
        if event.key() == Qt.Key_Up:
            statpan.panel.zoomIn()

        elif event.key() == Qt.Key_Down:
            statpan.panel.zoomOut()

        if key_enter:
            statpan.zoomEdited.emit(float(self.zoom.text()) / 100)

        QLineEdit.keyPressEvent(self.zoom, event)

def offsetGainKeyPressEvent(self, event):
    key_enter = (event.key() == Qt.Key_Return) or \
        (event.key() == Qt.Key_Enter)

    if key_enter:
        statpan = self.parent()
        statpan.offsetGainEdited.emit(statpan.offset.text(), statpan.gain.text(), statpan.gamma.text())

    QLineEdit.keyPressEvent(self, event)

class OpenImage(object):
    def __init__(self, imgpanel, path):
        self.imgpanel = imgpanel
        self.path = path

    def __call__(self):
        self.imgpanel.openImage(self.path)

class RecentMenu(QMenu):
    def __init__(self, parent=None):
        super().__init__('Recent', parent)
        self.imgpanel = self.parent()
        self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'images.png')))

    def showEvent(self, event):
        self.initactions()

    def initactions(self):
        self.clear()
        self.actions = []

        for rowid, timestamp, path in gui.qapp.history.yield_recent_paths():
            action = QAction(path, self)
            action.triggered.connect(OpenImage(self.imgpanel, path))
            self.addAction(action)
            self.actions.append(action)


class ImageViewerWidget(QWidget):
#class ImageViewerWidget(QtWidgets.QOpenGLWidget):
    #Image size seems to be limitted to 8192x8182
    #Tile shading limitation?

    pickerPositionChanged = Signal(int, int)
    zoomChanged = Signal(float)
    zoomPanChanged = Signal()
    pixelSelected = Signal(int, int)

    def __init__(self, parent=None):
        super().__init__(parent = parent)
        self.imgdata = ImageData()

        self.zoomValue = 1.0
        self.zoomPostScale = 1.0

        self.dispOffsetX = 0
        self.dispOffsetY = 0

        self.dragStartX = None
        self.dragStartY = None

        self.dispRoiStartX = self.dispOffsetX
        self.dispRoiStartY = self.dispOffsetY

        self.setBackgroundColor(*config['image background'])

        self.roi = SelRoiWidget(self)

        self.pickCursor = QCursor(QPixmap(str(respath / "icons" / "pixPick256.png")), 15, 15)
        self.dragCursor = QCursor(Qt.OpenHandCursor)

        self.setCursor(self.pickCursor)

        self.setMouseTracking(True)

        self.zoomPanChanged.connect(self.roi.recalcGeometry)

        self.refresh_title()

        self._scaledImage = None
        self.hqzoomout = config['image'].get('render_detail_hq', False)
        self.zoombind = config['image'].get('bind_zoom_absolute', False)

        self.push_selected_pixel = False

        self.setAcceptDrops(True)

    def setBackgroundColor(self, r, g, b):
        palette = self.palette()
        self.bgcolor = QColor(r,g,b)
        palette.setColor(self.backgroundRole(), self.bgcolor)
        self.setPalette(palette)
        #self.setAutoFillBackground(True)
        #This auto fill background seems not to work fine if
        #color is changed at runtime
        #As soon the parent of self is set to None, it retores back to the prior color
        #So it also happens after a relayout. (Even after a distribute)
        #There seems to be some cache of the prior background
        self.qpainter = QPainter()

    @property
    def vd(self):
        return self.imgdata

    def getImageCoordOfMouseEvent(self, event):
        x_float = event.pos().x() / self.zoomDisplay + self.dispOffsetX
        y_float = event.pos().y() / self.zoomDisplay + self.dispOffsetY
        #Round down
        return (int(x_float), int(y_float))

    def getImageCoordOfDisplayCenter(self):
        #get the current pixel position as seen on center of screen
        rect = self.geometry()
        centPointX = rect.width() // 2 / self.zoomDisplay + self.dispOffsetX
        centPointY = rect.height() // 2 / self.zoomDisplay + self.dispOffsetY
        return (centPointX, centPointY)

    def setZoomValue(self, value):
        self._scaledImage = None
        self._zoomValue = value

    def getZoomValue(self):
        return self._zoomValue

    def getZoomDisplay(self):
        return self._zoomValue * self.zoomPostScale

    zoomValue = property(getZoomValue, setZoomValue)
    zoomDisplay = property(getZoomDisplay)

    def setHigherZoomValue(self):
        if self.zoomValue in ZOOM_VALUES:
            i = ZOOM_VALUES.index(self.zoomValue)
            self.zoomValue = ZOOM_VALUES[min(i + 1, len(ZOOM_VALUES) - 1)]
            return self.zoomValue

        n = 0
        for zoomVal in ZOOM_VALUES:
            if zoomVal < self._zoomValue:
                n += 1
            else:
                self.zoomValue = ZOOM_VALUES[min(n, len(ZOOM_VALUES) - 1)]
                return self._zoomValue


        self.zoomValue = ZOOM_VALUES[0]
        return self.zoomValue

    def setLowerZoomValue(self):
        if self.zoomValue in ZOOM_VALUES:
            i = ZOOM_VALUES.index(self.zoomValue)
            self.zoomValue = ZOOM_VALUES[max(i - 1, 0)]
            return self.zoomValue

        n = 0
        for zoomVal in ZOOM_VALUES:
            if zoomVal < self.zoomValue:
                n += 1
            else:
                self.zoomValue = ZOOM_VALUES[max(n-1, 0)]
                return self.zoomValue

        self.zoomValue = ZOOM_VALUES[-1]
        return self.zoomValue

    def setClosestZoomValue(self):
        if self.zoomValue in ZOOM_VALUES:
            i = ZOOM_VALUES.index(self.zoomValue)
            self.zoomValue = ZOOM_VALUES[max(i - 1, 0)]
            return self.zoomValue

        n = 0
        for zoomVal in ZOOM_VALUES:
            if zoomVal < self.zoomValue:
                n += 1
            else:
                lower = ZOOM_VALUES[max(n-2, 0)]
                upper = ZOOM_VALUES[max(n-1, 0)]
                if (zoomVal - lower) < (upper - zoomVal):
                    self.zoomValue = lower
                else:
                    self.zoomValue = upper
                return self.zoomValue

        self.zoomValue = ZOOM_VALUES[-1]
        return self.zoomValue

    def setZoom(self, value=1, fixPointX = -1, fixPointY=-1):
        self.zoom(value, fixPointX, fixPointY, step=False)

    def zoomIn(self, fixPointX=-1, fixPointY=-1, fine=False):
        self.zoom(1, fixPointX, fixPointY, step=True, fine=fine)

    def zoomOut(self, fixPointX=-1, fixPointY=-1, fine=False):
        self.zoom(-1, fixPointX, fixPointY, step=True, fine=fine)

    def zoom(self, zoomValue=0, fixPointX=-1, fixPointY=-1, step=True, fine=False):
        self.dragStartX = None
        self.dragStartY = None

        if fixPointX == -1 or fixPointY == -1:
            (fixPointX, fixPointY) = self.getImageCoordOfDisplayCenter()

        tmpX = (fixPointX - self.dispOffsetX) * self.zoomDisplay
        tmpY = (fixPointY - self.dispOffsetY) * self.zoomDisplay

        if step == True:
            if fine:
                self.zoomValue = self.zoomValue + zoomValue * 0.01 * self.zoomValue
            else:
                if zoomValue > 0:
                    self.setHigherZoomValue()
                elif zoomValue < 0:
                    self.setLowerZoomValue()
        else:
            self.zoomValue = zoomValue

        # TO DO
        #Note that there is a rounding effect because the viewer will
        #Always round the top left corner to an integer
        #Start display a full pixel, not a part of a pixel
        #A lot of zoom ins and zoom outs will move the image to bottom, right
        #This effect also exists on the bigger steps but is less pronounced
        self.dispOffsetX = fixPointX - tmpX / self.zoomDisplay
        self.dispOffsetY = fixPointY - tmpY / self.zoomDisplay
        self.zoomPanChanged.emit()
        self.zoomChanged.emit(self.zoomValue)
        self.repaint()

        self.refresh_title()

    def refresh_title(self):
        #self.parent().setWindowTitle(f'Image Viewer {self.parent().id} - {self.zoomValue*100:.3g}%')
        pass

    def set_info_xy(self):
        self.parent().statuspanel.set_xy()

    def zoomAuto(self):
        if self.roi.isVisible():
            if not self.zoomToRoi():
                self.zoomFull()
        else:
            if not self.zoomFull():
                self.zoomFit()

    def zoomFull(self):
        """Zoom to the full image and do a best fit."""
        zoomRegionWidth = self.imgdata.qimg.width()
        zoomRegionHeight = self.imgdata.qimg.height()
        return self.zoomToRegion(0, 0, zoomRegionWidth, zoomRegionHeight, zoomSnap = False)

    def zoomFit(self):
        """Zoom to the full image and do a best fit. Snap of lower zoom value"""
        zoomRegionWidth = self.imgdata.qimg.width()
        zoomRegionHeight = self.imgdata.qimg.height()
        return self.zoomToRegion(0, 0, zoomRegionWidth, zoomRegionHeight, zoomSnap = True)

    def zoomToRoi(self):
        """Zoom to the region of interest and do a best fit."""
        zoomRegionX = self.roi.selroi.xr.start
        zoomRegionY = self.roi.selroi.yr.start
        zoomRegionWidth = self.roi.selroi.xr.stop - self.roi.selroi.xr.start
        zoomRegionHeight =  self.roi.selroi.yr.stop - self.roi.selroi.yr.start
        return self.zoomToRegion(zoomRegionX, zoomRegionY, zoomRegionWidth, zoomRegionHeight)

    def zoomNormalized(self, zoomRegionX, zoomRegionY, zoomRegionWidth, zoomRegionHeight, zoomSnap=True, emit=True, zoomValue=0):
        area = self.imgdata.width * self.imgdata.height
        zoomRegionX *= self.imgdata.width
        zoomRegionY *= self.imgdata.height
        zoomRegionWidth *= self.imgdata.width
        zoomRegionHeight *= self.imgdata.height
        self.zoomToRegion(zoomRegionX, zoomRegionY, zoomRegionWidth, zoomRegionHeight, zoomSnap, emit, zoomValue)

    def zoomToRegion(self, zoomRegionX, zoomRegionY, zoomRegionWidth, zoomRegionHeight, zoomSnap=True, emit=True, zoomValue=0):
        """Zoom to a certain region and do a best fit."""
        self.dragStartX = None
        self.dragStartY = None

        oldZoomValue = self.zoomValue
        oldDispOffsetX = self.dispOffsetX
        oldDispOffsetY = self.dispOffsetY

        xscale = self.width() / zoomRegionWidth
        yscale = self.height() / zoomRegionHeight

        if zoomValue == 0:

            self.zoomValue = min(xscale, yscale)

            if zoomSnap and (not self.zoomValue in ZOOM_VALUES):
                self.setLowerZoomValue()
                #self.setClosestZoomValue()

        else:
            self.zoomValue = zoomValue

        self.dispOffsetX = zoomRegionX - (self.width() / self.zoomDisplay - zoomRegionWidth) / 2
        self.dispOffsetY = zoomRegionY - (self.height() / self.zoomDisplay - zoomRegionHeight) / 2

        if (oldZoomValue != self.zoomValue) or \
            (oldDispOffsetX != self.dispOffsetX) or \
            (oldDispOffsetY != self.dispOffsetY):
            if emit:
                self.zoomPanChanged.emit()
            self.zoomChanged.emit(self.zoomValue)
            self.repaint()
            self.refresh_title()
            return True
        else:
            return False

    def visibleRegion(self, normalized=False, clip_square=False):
        x, y, w, h = self.dispOffsetX, self.dispOffsetY, self.width() / self.zoomDisplay, self.height() / self.zoomDisplay
        if clip_square:
            if h < w:
                x += (w - h) / 2
                w = h
            elif h > w:
                y += (h - w) / 2
                h = w

        if normalized:
            return x / self.imgdata.width, y / self.imgdata.height, w / self.imgdata.width, h / self.imgdata.height,
        return x, y, w, h

    def panned(self, manhattan=False):
        if (self.dragStartX is None) or (self.dragStartY is None):
            return None

        self.shiftX = (self.dragStartX - self.dragEndX) / self.zoomDisplay
        self.shiftY = (self.dragStartY - self.dragEndY) / self.zoomDisplay

        if manhattan:
            if abs(self.shiftX) < abs(self.shiftY):
                self.shiftX = 0
            else:
                self.shiftY = 0

        if self.shiftX != 0 or self.shiftY != 0:
            self.dispOffsetX = self.dispRoiStartX + self.shiftX
            self.dispOffsetY = self.dispRoiStartY + self.shiftY
            self.zoomPanChanged.emit()
            self.repaint()

    def mouseDoubleClickEvent(self, event):
        self.zoomAuto()

    def wheelEvent(self, event):
        if event.modifiers() & QtCore.Qt.ControlModifier:
            fine = True
        else:
            fine = False

        if event.delta() < 0:
            self.zoomOut(*self.getImageCoordOfMouseEvent(event), fine)
        else:
            self.zoomIn(*self.getImageCoordOfMouseEvent(event), fine)

    def mousePressEvent(self, event):
        #self.parent().setFocus()

        selectThisPanel(self)

        if event.buttons() == Qt.LeftButton or \
            (event.buttons() == Qt.MiddleButton):
            self.dragStartX = event.pos().x()
            self.dragStartY = event.pos().y()
            #roi value at the start of the dragging
            self.dispRoiStartX = self.dispOffsetX
            self.dispRoiStartY = self.dispOffsetY

        elif event.buttons() == Qt.RightButton:
            self.roiDragStartX, self.roiDragStartY = self.getImageCoordOfMouseEvent(event)

    def mouseMoveEvent(self, event):
        if (event.buttons() == Qt.LeftButton) or \
                (event.buttons() == Qt.MiddleButton):
            self.setCursor(self.dragCursor)
            self.dragEndX = event.pos().x()
            self.dragEndY = event.pos().y()
            self.panned(event.modifiers() & QtCore.Qt.ShiftModifier)

        elif (event.buttons() == Qt.RightButton):
            self.roiDragEndX, self.roiDragEndY = self.getImageCoordOfMouseEvent(event)
            self.roi.createState = True
            self.roi.setStartEndPoints(self.roiDragStartX, self.roiDragStartY, \
                self.roiDragEndX, self.roiDragEndY)
            self.roi.show()

        self.pickerPositionChanged.emit(*self.getImageCoordOfMouseEvent(event))

    def dragDistance(self, event):
        if self.dragStartX is None or self.dragStartY is None:
            return None

        return ((event.pos().x() - self.dragStartX)**2 + (event.pos().y() - self.dragStartY)**2) ** 0.5

    def mouseReleaseEvent(self, event):
        drag_distance = self.dragDistance(event)
        if not drag_distance is None and drag_distance <= 2:
            pixel_position = self.getImageCoordOfMouseEvent(event)
            if self.push_selected_pixel:
                panel = gui.qapp.panels.selected('console')
                panel.task.send_input(str(pixel_position))
                self.push_selected_pixel = False
            self.pixelSelected.emit(*pixel_position)

        self.setCursor(self.pickCursor)

        if self.roi.createState:
            self.roi.release_creation()

    def refresh(self):
        self._scaledImage = None
        self.repaint()

    def paintEvent(self, event):
        self.qpainter.begin(self)
        #logger.info(f'Painter: {self.qpainter.paintEngine().type()}')
        self.qpainter.fillRect(event.rect(), self.bgcolor)
        self.paintImage(self.qpainter)
        self.qpainter.end()

    def scaledImage(self):
        qimg = self.imgdata.qimg
        if self._scaledImage is None:
            self._scaledImage = qimg.scaledToWidth(int(qimg.width() * self.zoomDisplay), Qt.SmoothTransformation)
        return self._scaledImage

    def paintImage(self, qp, position=None):
        if position is None:
            sx = self.dispOffsetX
            sy = self.dispOffsetY
        else:
            sx, sy = position

        qp.setOpacity(1.0)

        if (self.zoomDisplay < 1) and (self.hqzoomout):
            qp.scale(1, 1)
            qp.drawImage(0, 0, self.scaledImage(), int(sx * self.zoomDisplay), int(sy * self.zoomDisplay), -1, -1)

        else:
            if config["image"].get('render_detail_smooth', False) and self.zoomDisplay < 1:
                qp.setRenderHint(qp.SmoothPixmapTransform)

            qp.scale(self.zoomDisplay, self.zoomDisplay)
            qp.translate(-sx, -sy)

            qp.drawImage(0, 0, self.imgdata.qimg, 0, 0, -1, -1)

            for layer in self.imgdata.layers.values():
                qp.setCompositionMode(layer['composition'])
                qp.drawImage(0, 0, layer['qimage'], 0, 0, -1, -1)

            qp.resetTransform()

    def dragEnterEvent(self, event):
        event.accept()

    def dropEvent(self, event):
        panel = thisPanel(self)
        mimeData = event.mimeData()

        dropedInFiles = []
        if mimeData.hasUrls():
            filenamelist = []

            for url in mimeData.urls():
                filename = url.toString(QUrl.FormattingOptions(QUrl.RemoveScheme)).replace('///','')
                dropedInFiles.append(filename)

        elif mimeData.hasText():

            filename = mimeData.text()
            dropedInFiles.append(filename)

        panel.openImage(dropedInFiles[0])

        for path in dropedInFiles[1:]:
            gui.img.open(path)


class ImageViewerBase(BasePanel):
    panelCategory = 'image'
    panelShortName = 'base'
    userVisible = False

    contentChanged = Signal(int, bool)
    gainChanged = Signal(int)
    visibleRegionChanged = Signal(float, float, float, float, bool, bool, float)
    roiChanged = Signal(int)

    classIconFile = str(respath / 'icons' / 'px16' / 'picture.png')

    def __init__(self, parent=None, panid=None):
        super().__init__(parent, panid, type(self).panelCategory)

        self.offset = 0
        self.gain = 1
        self.gamma = 1
        self.colormap = config['image color map']

        self.defaults = dict()
        self.defaults['offset'] = 0
        self.defaults['gain'] = 1
        self.defaults['gamma'] = 1

        self.createMenus()
        self.createStatusBar()

    def createMenus(self):
        self.fileMenu = self.menuBar().addMenu("&File")
        #self.editMenu = self.menuBar().addMenu("&Edit")
        self.editMenu = CheckMenu("&Edit", self.menuBar())
        self.menuBar().addMenu(self.editMenu)
        self.viewMenu = CheckMenu("&View", self.menuBar())
        self.menuBar().addMenu(self.viewMenu)
        self.selectMenu = self.menuBar().addMenu("&Select")
        self.canvasMenu = self.menuBar().addMenu("&Canvas")
        #self.imageMenu = self.menuBar().addMenu("&Image")
        self.imageMenu = CheckMenu("&Image", self.menuBar())
        self.processMenu = self.menuBar().addMenu("&Process")
        self.analyseMenu = self.menuBar().addMenu("&Analyse")

        ### File
        self.addMenuItem(self.fileMenu, 'New...'            , self.newImage,
            statusTip="Make a new image in this image viewer",
            icon = 'picture_empty.png')
        self.addMenuItem(self.fileMenu, 'Duplicate'         , self.duplicate,
            statusTip="Duplicate the image to a new image viewer",
            icon = 'application_double.png')
        self.addMenuItem(self.fileMenu, 'Open Image...' , self.openImageDialog,
            statusTip="Open an image",
            icon = 'folder_image.png')
        self.addMenuItem(self.fileMenu, 'Import Raw Image...', self.importRawImage,
            statusTip="Import Raw Image",
            icon = 'picture_go.png')
        self.fileMenu.addMenu(RecentMenu(self))
        self.addMenuItem(self.fileMenu, 'Save Image...' , self.saveImageDialog,
            statusTip="Save the image",
            icon = 'picture_save.png')
        self.addMenuItem(self.fileMenu, 'Close' , self.close_panel,
            statusTip="Close this image panel",
            icon = 'cross.png')

        ### Edit

        self.addMenuItem(self.editMenu, 'Show Prior Image', self.piorImage,
            enablecall = lambda: self.imviewer.imgdata.imghist.prior_length() > 0,
            statusTip="Get the prior image from the history stack and show it",
            icon = 'undo.png')
        self.addMenuItem(self.editMenu, 'Show Next Image', self.nextImage,
            enablecall = lambda: self.imviewer.imgdata.imghist.next_length() > 0,
            statusTip="Get the next image from the history stack and show it",
            icon = 'redo.png')

        self.editMenu.addSeparator()

        self.addMenuItem(self.editMenu, 'Copy 8bit Image to clipboard', self.placeRawOnClipboard,
            statusTip="Place the 8bit image on clipboard, offset and gain applied",
            icon = 'page_copy.png')
        self.addMenuItem(self.editMenu, 'Copy Display Image to clipboard', self.placeQimgOnClipboard,
            statusTip="Place the displayed image on clipboard with display processing applied",
            icon = 'page_copy.png')
        self.addMenuItem(self.editMenu, 'Paste into New Image', self.showFromClipboard,
            statusTip="Paste content of clipboard in this image viewer",
            icon = 'picture_clipboard.png')

        self.editMenu.addSeparator()

        ### View
        self.addMenuItem(self.viewMenu, 'Refresh', self.refresh,
            statusTip="Refresh the image",
            icon = 'update.png')
        self.addMenuItem(self.viewMenu, 'Zoom In' , self.zoomIn,
            statusTip="Zoom in 1 step",
            icon = 'zoom_in.png')
        self.addMenuItem(self.viewMenu, 'Zoom Out', self.zoomOut,
            statusTip="Zoom out 1 step",
            icon = 'zoom_out.png')

        zoomMenu = QMenu('Zoom')
        zoomMenu.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom.png')))
        self.viewMenu.addMenu(zoomMenu)
        self.addMenuItem(zoomMenu, 'Zoom 100%', self.setZoom100,
            statusTip="Zoom to a actual size (100%)",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_actual.png')))
        self.addMenuItem(zoomMenu, 'Zoom Fit'     , self.zoomFit,
            statusTip="Zoom to fit the image in the image viewer, snap on predefined zoom value",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_fit.png')))
        self.addMenuItem(zoomMenu, 'Zoom Full'    , self.zoomFull,
            statusTip="Zoom to fit the image in the image viewer",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_extend.png')))
        self.addMenuItem(zoomMenu, 'Zoom Auto'    , self.zoomAuto,
            statusTip="Toggle between to to selection and full image",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_refresh.png')))
        self.addMenuItem(zoomMenu, 'Zoom exact...'     , self.setZoom,
            statusTip="Zoom to a defined value",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'zoom_actual_equal.png')))

        self.viewMenu.addSeparator()

        self.addMenuItem(self.viewMenu, 'Default Offset && Gain', self.defaultOffsetGain,
            statusTip="Apply default offset, gain and gamma",
            icon=QtGui.QIcon(str(respath / 'icons' / 'px16' / 'unmark_to_download.png')))
        self.addMenuItem(self.viewMenu, 'Set Current as Default', self.setCurrentOffsetGainAsDefault,
            statusTip="Set the current offset, gain and gamma as default")
        self.addMenuItem(self.viewMenu, 'Offset && Gain...', self.offsetGainDialog,
            statusTip="Set offset and gain",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'weather_cloudy.png')))
        self.addMenuItem(self.viewMenu, 'Black && White...', self.blackWhiteDialog,
            statusTip="Set the black and white point",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'color_adjustment.png')))
        self.addMenuItem(self.viewMenu, 'Grey && Gain...', self.changeGreyGainDialog,
            statusTip="Set the mid grey level and gain",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast.png')))
        self.addMenuItem(self.viewMenu, 'Gain to Min-Max', self.gainToMinMax,
            statusTip="Auto level to min and max")
        self.gainSigmaMenu = QMenu('Gain to Sigma')
        self.viewMenu.addMenu(self.gainSigmaMenu)
        self.addMenuItem(self.gainSigmaMenu, 'Gain to Sigma 1', self.gainToSigma1)
        self.addMenuItem(self.gainSigmaMenu, 'Gain to Sigma 2', self.gainToSigma2)
        self.addMenuItem(self.gainSigmaMenu, 'Gain to Sigma 3', self.gainToSigma3)

        self.viewMenu.addSeparator()

        self.addMenuItem(self.viewMenu, 'HQ Zoom Out', self.toggle_hq,
            checkcall = lambda: self.imviewer.hqzoomout,
            statusTip = "Use high quality resampling on zoom levels < 100%")
            
        self.bindMenu = CheckMenu("Bind", self.viewMenu)
        self.addMenuItem(self.bindMenu, 'Bind All Image Viewers', self.bindImageViewers)
        self.addMenuItem(self.bindMenu, 'Unbind All Image Viewers', self.unbindImageViewers)        
        self.addMenuItem(self.bindMenu, 'Absolute Zoom Link', self.toggle_zoombind,
            checkcall = lambda: self.imviewer.zoombind,
            statusTip = "If binded to other image viewer, bind with absolute zoom value")        
        
        self.addMenuItem(self.viewMenu, 'Colormap...'    , self.setColorMap,
            statusTip="Set the color map for monochroom images",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'dopplr.png')))
        self.addMenuItem(self.viewMenu, 'Background Color...'    , self.setBackground,
            statusTip="Set the background color...",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'document_background.png')))
        self.addMenuItem(self.viewMenu, 'Selection Color...'    , self.setRoiColor,
            statusTip="Set the Selection color...",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'color_swatch.png')))

        self.chooseValFormat = QMenu('Value Format')
        self.chooseValFormat.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'pilcrow.png')))
        self.chooseValFormat.addAction(QAction("Decimal", self, triggered=lambda: self.statuspanel.set_val_format('dec')))
        self.chooseValFormat.addAction(QAction("Hex", self, triggered=lambda: self.statuspanel.set_val_format('hex')))
        self.chooseValFormat.addAction(QAction("Binary", self, triggered=lambda: self.statuspanel.set_val_format('bin')))
        self.viewMenu.addMenu(self.chooseValFormat)

        ### Select
        self.addMenuItem(self.selectMenu, 'Select Full Image', self.selectAll,
            statusTip="Select Full Image")
        self.addMenuItem(self.selectMenu, 'Deselect', self.selectNone,
            statusTip="Deselect, select nothing")
        self.addMenuItem(self.selectMenu, 'Select dialog...', self.setRoi,
            statusTip="Select with input numbers dialog",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'region_of_interest.png')))
        self.addMenuItem(self.selectMenu, 'Jump to Coordinates'   , self.jumpToDialog,
            statusTip="Select 1 pixel and zoom to it",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'canvas.png')))
        self.addMenuItem(self.selectMenu, 'Mask Value...'   , self.maskValue,
            statusTip="Mask pixels based on value",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'find.png')))

        ### Canvas
        self.addMenuItem(self.canvasMenu, 'Flip Horizontal', self.flipHorizontal,
            statusTip="Flip the image Horizontal",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'shape_flip_horizontal.png')))
        self.addMenuItem(self.canvasMenu, 'Flip Vertical'  , self.flipVertical,
            statusTip="Flip the image Vertical",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'shape_flip_vertical.png')))
        self.addMenuItem(self.canvasMenu, 'Rotate Left 90' , self.rotate90,
            statusTip="Rotate the image 90 degree anti clockwise",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'shape_rotate_anticlockwise.png')))
        self.addMenuItem(self.canvasMenu, 'Rotate Right 90', self.rotate270,
            statusTip="Rotate the image 90 degree clockwise",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'shape_rotate_clockwise.png')))
        self.addMenuItem(self.canvasMenu, 'Rotate 180'     , self.rotate180,
            statusTip="Rotate the image 180 degree")
        self.addMenuItem(self.canvasMenu, 'Rotate any Angle...', triggered=self.rotateAny, enabled=has_scipy,
            statusTip="Rotate any angle")
        self.addMenuItem(self.canvasMenu, 'Crop on Selection', self.crop,
            statusTip="Crop the image on the current rectangle selection",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'transform_crop.png')))
        self.addMenuItem(self.canvasMenu, 'Resize Canvas...', self.canvasResize,
            statusTip="Add or remove borders",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'canvas_size.png')))
        self.addMenuItem(self.canvasMenu, 'Resize Image', triggered=self.resize, enabled=has_scipy,
            statusTip="Resize the image by resampling",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'resize_picture.png')))

        ### Image
        self.addMenuItem(self.imageMenu, 'Swap RGB | BGR', self.swapRGB,
            statusTip="Swap the blue with red channel",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'color.png')))
        self.addMenuItem(self.imageMenu, 'to Monochroom', self.toMonochroom,
            statusTip="Convert an RGB image to monochroom grey",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'convert_color_to_gray.png')))
        self.addMenuItem(self.imageMenu, 'to Photometric Monochroom', self.toPhotoMonochroom,
            statusTip="Convert an RGB image to photometric monochroom grey",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'convert_color_to_gray.png')))
        self.addMenuItem(self.imageMenu, 'to 8-bit', self.to8bit,
            enablecall = self.is16bit)
        self.addMenuItem(self.imageMenu, 'to 16-bit', self.to16bit,
            enablecall = self.is8bit)
        self.addMenuItem(self.imageMenu, 'Fill...'          , self.fillValue,
            statusTip="Fill the image with the same value",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'paintcan.png')))
        self.addMenuItem(self.imageMenu, 'Add noise...'     , self.addNoise,
            statusTip="Add Gaussian noise")
        self.addMenuItem(self.imageMenu, 'Invert', self.invert,
            statusTip="Invert the image")
        self.addMenuItem(self.imageMenu, 'Adjust Lighting...', self.adjustLighting,
            statusTip="Adjust the pixel values",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'contrast.png')))
        self.addMenuItem(self.imageMenu, 'Adjust Gamma...', self.adjustGamma,
            statusTip="Adjust the gamma")

        #Process
        self.addMenuItem(self.processMenu, 'Bayer Split', self.bayer_split_tiles,
            statusTip="Split to 4 images based on the Bayer kernel",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'pictures_thumbs.png')))
        self.addMenuItem(self.processMenu, 'Colored Bayer', self.colored_bayer)
        self.addMenuItem(self.processMenu, 'Demosaic', self.demosaic, enabled=has_scipy,
            statusTip="Demosaic",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'things_digital.png')))
        self.addMenuItem(self.processMenu, 'Make Blueprint', self.makeBlueprint,
            statusTip="Make a thumbnail (8x smaller) with blowup high frequencies",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'map_blue.png')))

        #Analyse
        self.addMenuItem(self.analyseMenu, 'Horizontal Spectrogram', self.horizontalSpectrogram,
            statusTip="Horizontal Spectrogram")
        self.addMenuItem(self.analyseMenu, 'Vertical Spectrogram', self.verticalSpectrogram,
            statusTip="Vertical Spectrogram")
        self.addMenuItem(self.analyseMenu, 'Measure Distance', self.measureDistance,
            statusTip="Measure Distance",
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / 'geolocation_sight.png')))

        self.addBaseMenu(['levels', 'values', 'image'])

    def get_select_menu(self):
        #The select menu should be index 4 from the menuBar children
        return self.menuBar().children()[4]

    def createStatusBar(self):
        self.statuspanel = StatusPanel(self)
        self.statusBar().addWidget(self.statuspanel)

    def set_info_xy_val(self, x, y):
        try:
            val = self.imviewer.imgdata.statarr[y, x]

        except:
            val = None

        self.statuspanel.set_xy_val(x, y, val)

    def addBindingTo(self, category, panid):
        targetPanel = super().addBindingTo(category, panid)
        if targetPanel is None: return None
        if targetPanel.category == 'image':
            self.visibleRegionChanged.connect(targetPanel.changeVisibleRegion)
        elif targetPanel.category == 'levels':
            self.contentChanged.connect(targetPanel.imageContentChanged)
            self.roiChanged.connect(targetPanel.roiChanged)
            self.gainChanged.connect(targetPanel.imageGainChanged)
        elif targetPanel.category == 'values':
            self.imviewer.pixelSelected.connect(targetPanel.pick)
        return targetPanel

    def removeBindingTo(self, category, panid):
        targetPanel = super().removeBindingTo(category, panid)
        if targetPanel is None: return None
        if targetPanel.category == 'image':
            self.visibleRegionChanged.disconnect(targetPanel.changeVisibleRegion)
        elif targetPanel.category == 'levels':
            self.contentChanged.disconnect(targetPanel.imageContentChanged)
            self.gainChanged.disconnect(targetPanel.imageGainChanged)
        elif targetPanel.category == 'values':
            self.imviewer.pixelSelected.disconnect(targetPanel.pick)
        return targetPanel

    def changeVisibleRegion(self, x, y, w, h, zoomSnap, emit, zoomValue):
        self.imviewer.zoomNormalized(x, y, w, h, zoomSnap, emit, zoomValue)
        self.imviewer.roi.recalcGeometry()

    ############################
    # File Menu Connections
    def newImage(self):

        with ActionArguments(self) as args:
            args['width'] = 1920*2
            args['height'] = 1080*2
            args['channels'] = 1
            args['dtype'] = 'uint8'
            args['mean'] = 128

        if args.isNotSet():
            dtypes = ['uint8', 'uint16', 'double']

            options_form = [('Width', args['width']),
                       ('Height', args['height']),
                       ('Channels', args['channels']),
                       ('dtype', [1] + dtypes),
                       ('mean', args['mean'])]

            result = fedit(options_form)
            if result is None: return
            args['width'], args['height'], args['channels'], dtype_ind, args['mean'] = result
            args['dtype'] = dtypes[dtype_ind-1]

        shape = [args['height'], args['width']]
        if args['channels'] > 1: shape = shape + [args['channels']]

        arr = np.ndarray(shape, args['dtype'])
        arr[:] = args['mean']

        self.show_array(arr, zoomFitHist=True)

    def duplicate(self, floating=False):
        newPanel = super().duplicate(floating)
        newPanel.show_array(self.ndarray)
        return newPanel

    def openImageDialog(self):
        filepath = here / 'images' / 'default.png'

        with ActionArguments(self) as args:
            args['filepath'] = here / 'images' / 'default.png'
            args['format'] = None

        if args.isNotSet():
            if has_imafio:
                args['filepath'], filter = gui.getfile(filter=IMAFIO_QT_READ_FILTERS, title='Open Image File (Imafio)', file=str(args['filepath']))
                if args['filepath'] == '': return
                args['format'] = FILTERS_NAMES[filter]

            else:
                args['filepath'], filter = gui.getfile(title='Open Image File (PIL)', file=str(args['filepath']))
                args['format'] = None
                if args['filepath'] == '': return

        self.openImage(args['filepath'], args['format'])

    def openImage(self, filepath, format=None):
        if has_imafio:
            arr = self.openImageImafio(filepath, format)
        else:
            arr = self.openImagePIL(filepath)

        if not arr is None:
            self.long_title = str(filepath)
            gui.qapp.history.storepath(str(filepath))
            self.show_array(arr, zoomFitHist=True)
            self.zoomFull()

    def openImagePIL(self, filepath):
        with gui.qapp.waitCursor(f'Opening image using PIL {filepath}'):
            from PIL import Image
            logger.info(f'Using PIL library')
            image = Image.open(str(filepath))
            arr = np.array(image)
        return arr

    def openImageImafio(self, filepath, format=None):
        with gui.qapp.waitCursor(f'Opening image using imageio {filepath} {format}'):
            FormatClass = imageio.formats[str(filepath)]
            logger.info(f'Using FormatClass {repr(FormatClass)}')
            arr = imageio.imread(str(filepath), format=format)
        return arr

    def importRawImage(self):
        import struct

        filepath = here / 'images' / 'default.png'
        filepath = gui.getfile(file=str(filepath))[0]
        if filepath == '': return

        fp = open(filepath, 'br')
        data = fp.read()
        fp.close()

        #somehwhere in the header, there is the resolution
        #image studio: 128 bytes header, 4 bytes=width, 4 bytes=height, 120 bytes=???
        header = 128
        dtype = 'uint16'
        width = struct.unpack('<I', data[0:4])[0]
        height = struct.unpack('<I', data[4:8])[0]

        form = [('data offset', header),
            ('dtype', 'uint16'),
            ('width',  width),
            ('height', height)]

        returns = fedit(form, title="raw parameters")

        offset = returns[0]
        dtype = returns[1]
        width = returns[2]
        height = returns[3]

        with gui.qapp.waitCursor():
            dtype = np.dtype(dtype)

            leftover = len(data) - (width * height  * dtype.itemsize + offset)

            if leftover > 0:
                print('Too much data found (%d bytes too many)' % leftover)

            elif leftover < 0:
                print('Not enough data found (missing %d bytes)' % (-leftover))

            arr = np.ndarray(shape=(height, width), dtype=dtype, buffer=data[offset:])
            self.show_array(arr, zoomFitHist=True)
            self.zoomFull()
            gui.qapp.history.storepath(str(filepath))

    def saveImageDialog(self):
        if has_imafio:
            filepath, filter = gui.putfile(filter=IMAFIO_QT_WRITE_FILTERS, title='Save Image using Imafio',
                                    defaultfilter=IMAFIO_QT_WRITE_FILTER_DEFAULT)
            if filepath == '': return
            format = FILTERS_NAMES[filter]
            self.saveImage(filepath, format)
        else:
            filepath, filter = gui.putfile(title='Save Image using PIL')
            if filepath == '': return
            self.saveImage(filepath)

    def saveImage(self, filepath, format=None):
        if has_imafio:
            self.saveImageImafio(filepath, format)
        else:
            self.saveImagePIL(filepath)

        gui.qapp.history.storepath(str(filepath))

    def saveImagePIL(self, filepath):
        with gui.qapp.waitCursor():
            from PIL import Image

            image = Image.fromarray(self.ndarray)
            image.save(str(filepath))

    def saveImageImafio(self, filepath, format):
        if format is None:
            from imageio.core import Request
            format = imageio.formats.search_write_format(Request(filepath, 'wi')).name

        if format == 'JPEG-FI':
            (quality, progressive, optimize, baseline) = gui.fedit([('quality', 90), ('progressive', False), ('optimize', False), ('baseline', False)])

            with gui.qapp.waitCursor(f'Saving to {filepath}'):
                imageio.imwrite(filepath, self.ndarray, format,
                    quality=quality, progressive=progressive,
                    optimize=optimize, baseline=baseline)

        elif format == 'TIFF-FI':
            compression_options = {
                'none': imageio.plugins.freeimage.IO_FLAGS.TIFF_NONE,
                'default': imageio.plugins.freeimage.IO_FLAGS.TIFF_DEFAULT,
                'packbits': imageio.plugins.freeimage.IO_FLAGS.TIFF_PACKBITS,
                'adobe': imageio.plugins.freeimage.IO_FLAGS.TIFF_ADOBE_DEFLATE,
                'lzw': imageio.plugins.freeimage.IO_FLAGS.TIFF_LZW,
                'deflate': imageio.plugins.freeimage.IO_FLAGS.TIFF_DEFLATE,
                'logluv': imageio.plugins.freeimage.IO_FLAGS.TIFF_LOGLUV}
            (compression_index,) = gui.fedit([('compression', [2] + list(compression_options.keys()))])
            compression = list(compression_options.keys())[compression_index-1]
            compression_flag = compression_options[compression]

            with gui.qapp.waitCursor(f'Saving to {filepath}'):
                imageio.imwrite(filepath, self.ndarray, format, flags=compression_flag)

        elif format == 'PNG-FI':
            compression_options = [('None', 0), ('Best Speed', 1), ('Default', 6), ('Best Compression', 9)]
            (compression_index, quantize, interlaced) = gui.fedit([('compression', [2] + [item[0] for item in compression_options]), ('quantize', 0), ('interlaced', True)])
            compression = compression_options[compression_index-1][1]

            print(f'compression: {compression}')

            with gui.qapp.waitCursor(f'Saving to {filepath}'):
                imageio.imwrite(filepath, self.ndarray, format, compression=compression, quantize=quantize, interlaced=interlaced)

        elif format == 'PNG-PIL':
            compression_options = [('None', 0), ('Best Speed', 1), ('Default', 6), ('Best Compression', 9)]
            (compression_index, quantize, optimize) = gui.fedit([('compression', [4] + [item[0] for item in compression_options]), ('quantize', 0), ('optimize', True)])
            compression = compression_options[compression_index-1][1]
            if quantize == 0: quantize = None

            print(f'compression: {compression}')

            with gui.qapp.waitCursor(f'Saving to {filepath}'):
                imageio.imwrite(filepath, self.ndarray, format, compression=compression, quantize=quantize, optimize=optimize)

        else:
            with gui.qapp.waitCursor(f'Saving to {filepath}'):
                imageio.imwrite(filepath, self.ndarray, format)

    def close_panel(self):
        super().close_panel()

        #Deleting self.imviewer doesn't seem to delete the imgdata
        del self.imviewer.imgdata


    ############################
    # Edit Menu Connections

    def piorImage(self):
        if self.imviewer.imgdata.imghist.prior_length() > 0:
            arr = self.imviewer.imgdata.imghist.prior(self.ndarray)
            self.imviewer.imgdata.sharray = None
            self.show_array(arr)

    def nextImage(self):
        if self.imviewer.imgdata.imghist.next_length() > 0:
            arr = self.imviewer.imgdata.imghist.next(self.ndarray)
            self.imviewer.imgdata.sharray = None
            self.show_array(arr)

    #---------------------------

    def placeRawOnClipboard(self):
        clipboard = self.qapp.clipboard()
        array = self.ndarray
        qimg = imconvert.process_ndarray_to_qimage_8bit(array, 0, 1)
        clipboard.setImage(qimg)

    def placeQimgOnClipboard(self):
        clipboard = self.qapp.clipboard()
        #If qimg is not copied, GH crashes on paste after the qimg instance has been garbaged!
        #Clipboard can only take ownership if the object is a local?
        qimg = self.imviewer.imgdata.qimg.copy()
        clipboard.setImage(qimg)

    def showFromClipboard(self):
        arr = gui.get_clipboard_image()
        self.show_array(arr)

    ############################
    # View Menu Connections

    def refresh(self):
        with gui.qapp.waitCursor(f'Refreshing {self.short_title}'):
            self.show_array(None)

    @property
    def white(self):
        gain1_range = self.imviewer.imgdata.get_natural_range()
        return gain1_range / self.gain + self.offset

    def offsetGainDialog(self):

        with ActionArguments(self) as args:
            args['offset'] = self.offset
            args['gain'] = self.gain
            args['gamma'] = self.gamma
            args['cmap'] = self.colormap

        if args.isNotSet():
            colormaps = imconvert.colormaps
            cmapind = colormaps.index(self.colormap) + 1

            form = [('Offset', self.offset * 1.0),
                    ('Gain', self.gain * 1.0),
                    ('Gamma', self.gamma * 1.0),
                    ('Color Map', [cmapind] + colormaps)]

            results = fedit(form)
            if results is None: return
            self.offset, self.gain, self.gamma, cmapind = results
            self.colormap = colormaps[cmapind-1]

        else:
            self.offset, self.gain = args['offset'], args['gain']
            self.gamma, self.colormap = args['gamma'], args['cmap']

        self.refresh_offset_gain()


    def setCurrentOffsetGainAsDefault(self):
        self.defaults['offset'] = self.offset
        self.defaults['gain'] = self.gain
        self.defaults['gamma'] = self.gamma


    def defaultOffsetGain(self):
        offset = self.defaults['offset']
        gain = self.defaults['gain']
        gamma = self.defaults['gamma']
        self.changeOffsetGain(offset, gain, gamma)


    def changeOffsetGain(self, offset, gain, gamma):
        if isinstance(offset, str):
            if offset == 'default':
                offset = self.defaults['offset']
            else:
                offset = eval(offset)
        if isinstance(gain, str):
            if gain == 'default':
                gain = self.defaults['gain']
            else:
                gain = eval(gain)
        if isinstance(gamma, str):
            if gamma == 'default':
                gamma = self.defaults['gamma']
            else:
                gamma = eval(gamma)

        if not offset is None: self.offset = offset
        if not gain is None: self.gain = gain
        if not gamma is None: self.gamma = gamma
        self.refresh_offset_gain()

    def blackWhiteDialog(self):

        with ActionArguments(self) as args:
            args['black'] = self.offset
            args['white'] = self.white
            args['cmap'] = self.colormap

        if args.isNotSet():
            colormaps = imconvert.colormaps
            cmapind = colormaps.index(self.colormap) + 1

            black = self.offset
            gain1_range = self.imviewer.imgdata.get_natural_range()

            form = [('Black', black),
                    ('White', self.white),
                    ('Color Map', [cmapind] + colormaps)]

            results = fedit(form)
            if results is None: return
            black, white, cmapind = results
            self.colormap = colormaps[cmapind-1]

        else:
            black, white = args['black'], args['white']
            self.colormap  = args['cmap']

        self.changeBlackWhite(black, white)

    def changeBlackWhite(self, black, white):
        if isinstance(black, str):
            black = eval(black)
        if isinstance(white, str):
            white = eval(white)

        if black == white:
            print(f'Warning: black and white are set the same ({black}). Setting to mid grey!')
            self.changeMidGrey(black)
            return

        gain1_range = self.imviewer.imgdata.get_natural_range()

        if not (black is None or white is None):
            self.offset = black
            self.gain = gain1_range / (white - black)
        elif white is None:
            white = self.white
            self.offset = black
            self.gain = gain1_range / (white - self.offset)
        elif black is None:
            self.gain = gain1_range / (white - self.offset)

        self.refresh_offset_gain()

    def changeGreyGainDialog(self):

        gain1_range = self.imviewer.imgdata.get_natural_range()
        grey = self.offset + gain1_range / self.gain / 2

        with ActionArguments(self) as args:
            args['grey'] = grey
            args['gain'] = self.gain
            args['cmap'] = self.colormap

        if args.isNotSet():
            colormaps = imconvert.colormaps
            cmapind = colormaps.index(self.colormap) + 1

            form = [('Grey', grey),
                    ('Gain', self.gain * 1.0),
                    ('Color Map', [cmapind] + colormaps)]

            results = fedit(form)
            if results is None: return
            grey, gain, cmapind = results
            self.colormap = colormaps[cmapind-1]

        else:
            grey = args['grey']
            gain = args['gain']
            self.colormap = args['cmap']

        self.changeMidGrey(grey, gain)

    def changeMidGrey(self, midgrey, gain=None):
        if not gain is None: self.gain = gain
        gain1_range = self.imviewer.imgdata.get_natural_range()
        self.offset = midgrey - gain1_range / self.gain / 2
        self.refresh_offset_gain()

    def gainToMinMax(self):
        black = self.ndarray.min()
        white = self.ndarray.max()
        self.changeBlackWhite(black, white)

    def gainToSigma1(self):
        with gui.qapp.waitCursor('Gain 1 sigma'):
            self.gainToSigma(1)

    def gainToSigma2(self):
        with gui.qapp.waitCursor('Gain 2 sigma'):
            self.gainToSigma(2)

    def gainToSigma3(self):
        with gui.qapp.waitCursor('Gain 3 sigma'):
            self.gainToSigma(3)

    def gainToSigma(self, sigma=3, roi=None):
        chanstats = self.imviewer.imgdata.chanstats

        if roi is None:
            roi = self.imviewer.roi.isVisible()

        elif roi and not self.imviewer.roi.isVisible():
            roi = False

        if roi:
            clrs = set(('RK','RR', 'RG', 'RB'))
        else:
            clrs = set(('K', 'R', 'G', 'B'))

        clrs = clrs.intersection(set(chanstats.keys()))

        blacks = dict()
        whites = dict()
        for clr in clrs:
            stats = chanstats[clr]
            if stats.arr2d is None: continue
            hist = stats.histogram(1)
            starts = stats.starts(1)
            blacks[clr], whites[clr] = get_sigma_range_for_hist(starts, hist, sigma)

        black = min(blacks.values())
        white = max(whites.values())

        if self.ndarray.dtype in ['uint8', 'uint16']:
            white += 1

        self.changeBlackWhite(black, white)

    def zoomIn(self):
        self.imviewer.zoomIn()

    def zoomOut(self):
        self.imviewer.zoomOut()

    def setZoom100(self):
        self.imviewer.setZoom(1)

    def setZoom(self):
        with ActionArguments(self) as args:
            args['zoom'] = self.imviewer.zoomValue * 100

        if args.isNotSet():
            results = fedit([('Zoom value %', args['zoom'])])
            if results is None: return
            args['zoom'] = results[0]

        self.imviewer.setZoom(args['zoom'] / 100)

    def setZoomValue(self, value):
        self.imviewer.setZoom(value)

    def zoomFit(self):
        self.imviewer.zoomFit()

    def zoomFull(self):
        self.imviewer.zoomFull()

    def zoomAuto(self):
        self.imviewer.zoomAuto()

    def setColorMap(self):
        with ActionArguments(self) as args:
            args['cmap'] = 'grey'

        if args.isNotSet():
            colormapdialog = ColorMapDialog()
            colormapdialog.exec_()
            self.colormap = colormapdialog.cm_name
        else:
            self.colormap = args['cmap']

        self.refresh_offset_gain()

    def toggle_hq(self):
        self.imviewer.hqzoomout = not self.imviewer.hqzoomout
        self.show_array(None)

    def toggle_zoombind(self):
        self.imviewer.zoombind = not self.imviewer.zoombind
        
    def bindImageViewers(self):
        for src_panid, src_panel in gui.qapp.panels['image'].items():
            for tgt_panid, tgt_panel in gui.qapp.panels['image'].items():            
                if src_panid == tgt_panid: continue
                src_panel.addBindingTo('image', tgt_panid)
                
    def unbindImageViewers(self):
        for src_panid, src_panel in gui.qapp.panels['image'].items():
            for tgt_panid, tgt_panel in gui.qapp.panels['image'].items():            
                if src_panid == tgt_panid: continue
                src_panel.removeBindingTo('image', tgt_panid)                

    def setBackground(self):

        old_color = self.imviewer.palette().window().color()
        rgb = old_color.toTuple()[:3]

        with ActionArguments(self) as args:
            args['r'] = rgb[0]
            args['g'] = rgb[1]
            args['b'] = rgb[2]

        if args.isNotSet():
            color = QColorDialog.getColor(old_color)

            try:
                rgb = color.toTuple()[:3]
            except:
                rgb = (0,0,0)

        else:
            rgb = (args['r'], args['g'], args['b'])

        config['image background'] = rgb
        self.imviewer.setBackgroundColor(*config['image background'])


    def setRoiColor(self):
        
        old_color = QtGui.QColor(*config['roi color'])
        rgb = old_color.toTuple()[:3]
        
        with ActionArguments(self) as args:
            args['r'] = rgb[0]
            args['g'] = rgb[1]
            args['b'] = rgb[2]
            
        if args.isNotSet():
            color = QColorDialog.getColor(old_color)

            try:
                rgb = color.toTuple()[:3]
            except:
                rgb = (0,0,0)
                
        else:
            rgb = (args['r'], args['g'], args['b'])                
            
        config['roi color'] = rgb
        self.imviewer.roi.initUI()

    ############################
    # Select Menu Connections

    def selectAll(self):
        selroi = self.imviewer.imgdata.selroi
        selroi.reset()
        self.imviewer.roi.clip()
        self.imviewer.roi.show()

    def selectNone(self):
        self.imviewer.imgdata.selroi.reset()
        self.imviewer.roi.clip()
        self.imviewer.roi.hide()

    def setRoi(self):
        selroi = self.imviewer.imgdata.selroi

        form = [('x start', selroi.xr.start),
                ('x stop', selroi.xr.stop),
                ('x step', selroi.xr.step),
                ('y start', selroi.yr.start),
                ('y stop', selroi.yr.stop),
                ('y step', selroi.yr.step)]

        r = fedit(form, title='ROI')
        if r is None: return

        selroi.xr.start = r[0]
        selroi.xr.stop = r[1]
        selroi.xr.step = r[2]
        selroi.yr.start = r[3]
        selroi.yr.stop = r[4]
        selroi.yr.step = r[5]

        self.imviewer.roi.clip()
        self.imviewer.roi.show()

    def jumpToDialog(self):
        selroi = self.imviewer.imgdata.selroi

        form = [('x', selroi.xr.start),
                ('y', selroi.yr.start)]

        results = fedit(form, title='Position')
        if results is None: return
        x, y = results
        self.jumpTo(x, y)

    def jumpTo(self, x, y):
        selroi = self.imviewer.imgdata.selroi

        selroi.xr.start, selroi.yr.start = x, y
        selroi.xr.stop = selroi.xr.start + 1
        selroi.xr.step = 1
        selroi.yr.stop = selroi.yr.start + 1
        selroi.yr.step = 1

        self.imviewer.roi.clip()
        self.imviewer.roi.show()
        self.imviewer.zoomToRoi()
        self.roiChanged.emit(self.panid)

    def maskValue(self):
        array = self.ndarray
        evalOptions = ['Equal', 'Smaller', 'Larger']

        if array.ndim == 2:
            form = [('Evaluate', [1] + evalOptions),
                    ('Value', 0)]
        elif array.ndim == 3:
            form = [('Evaluate', [1] + evalOptions),
                ('Red', 0),
                ('Green', 0),
                ('Blue', 0)]

        result = fedit(form, title='Mask')

        if result is None:
            self.imviewer.imgdata.set_mask(None)
            self.imviewer.refresh()
            return

        evalind, *values = result

        if evalind == 1:
            if array.ndim == 2:
                mask = (array == values[0])

            elif array.ndim == 3:
                mask0 = (array[:,:,0] == values[0])
                mask1 = (array[:,:,1] == values[1])
                mask2 = (array[:,:,2] == values[2])
                mask = mask0 & mask1 & mask2

        elif evalind == 2:
            if array.ndim == 2:
                mask = (array < values[0])

            elif array.ndim == 3:
                mask0 = (array[:,:,0] < values[0])
                mask1 = (array[:,:,1] < values[1])
                mask2 = (array[:,:,2] < values[2])
                mask = mask0 & mask1 & mask2

        elif evalind == 3:
            if array.ndim == 2:
                mask = (array > values[0])

            elif array.ndim == 3:
                mask0 = (array[:,:,0] > values[0])
                mask1 = (array[:,:,1] > values[1])
                mask2 = (array[:,:,2] > values[2])
                mask = mask0 & mask1 & mask2

        self.imviewer.imgdata.set_mask(mask)
        self.imviewer.refresh()

    ############################
    # Canvas Menu Connections

    def flipHorizontal(self):
        self.show_array(self.ndarray[:, ::-1])


    def flipVertical(self):
        self.show_array(self.ndarray[::-1, :])


    def rotate90(self):
        rotated = np.rot90(self.ndarray, 1).copy()
        self.show_array(rotated)


    def rotate180(self):
        self.show_array(self.ndarray[::-1, ::-1])


    def rotate270(self):
        rotated = np.rot90(self.ndarray, 3).copy()
        self.show_array(rotated)


    def rotateAny(self):
        with ActionArguments(self) as args:
            args['angle'] = 0.0

        if args.isNotSet():
            form = [('Angle', args['angle'])]
            results = fedit(form)
            if results is None: return
            args['angle'] = results[0]

        with gui.qapp.waitCursor(f'Rotating {args["angle"]} degree'):
            procarr = scipy.ndimage.rotate(self.ndarray, args['angle'], reshape=True)
            self.show_array(procarr)


    def crop(self):
        self.select()
        croped_array = gui.vr.copy()
        gui.img.show(croped_array)
        self.selectNone()


    def canvasResize(self):
        old_height, old_width = self.ndarray.shape[:2]

        with ActionArguments(self) as args:
            args['width'], args['height'] = old_width, old_height

        channels = self.ndarray.shape[2] if self.ndarray.ndim == 3 else 1

        if args.isNotSet():

            form = [('Width', args['width']), ('Height', args['height'])]
            results = fedit(form)
            if results is None: return
            args['width'], args['height'] = results

        new_width = args['width']
        new_height = args['height']

        if channels == 1:
            procarr = np.ndarray((new_height, new_width), dtype=self.ndarray.dtype)
        else:
            procarr = np.ndarray((new_height, new_width, channels), dtype=self.ndarray.dtype)

        #What with the alpha channel?
        procarr[:] = 0

        width = min(old_width, new_width)
        height = min(old_height, new_height)
        ofow = (old_width - width) // 2
        ofnw = (new_width - width) // 2
        ofoh = (old_height - height) // 2
        ofnh = (new_height - height) // 2
        procarr[ofnh:ofnh+height, ofnw:ofnw+width, ...] = self.ndarray[ofoh:ofoh+height, ofow:ofow+width, ...]
        self.show_array(procarr)


    def resize(self):
        source = self.sharray.ndarray
        shape = self.sharray.shape

        form = [("width", shape[1]), ("height", shape[0]), ("order", 1)]
        results = fedit(form)
        if results is None: return
        width, height, order = results

        factorx = width / shape[1]
        factory = height / shape[0]

        if source.ndim == 2:
            scaled = scipy.ndimage.zoom(source, (factory, factorx), order=order, mode="nearest")

        elif source.ndim == 3:
            #some bug here
            scaled = scipy.ndimage.zoom(source, (factory, factorx, 1.0), order=order, mode="nearest")
            #returned array dimensions are not on the expected index

        self.show_array(scaled)


    ############################
    # Image Menu Connections

    def fillValue(self):
        """
        :param float value:
        """
        with ActionArguments(self) as args:
            args['value'] = 0.0

        if args.isNotSet():
            form = [('Value', args['value'])]
            results = fedit(form)
            if results is None: return
            args['value'] = results[0]

        procarr = self.ndarray.copy()
        procarr[:] = args['value']
        self.show_array(procarr)


    def addNoise(self):
        form = [('Standard Deviation', 1.0)]
        results = fedit(form)
        if results is None: return
        std = float(results[0])

        shape = self.ndarray.shape
        dtype = self.ndarray.dtype
        procarr = clip_array(self.ndarray + np.random.randn(*shape) * std + 0.5, dtype)
        self.show_array(procarr)


    def invert(self):
        procarr =  ~self.ndarray
        self.show_array(procarr)


    def swapRGB(self):
        if not self.ndarray.ndim >= 3:
            gui.dialog.msgbox('The image has not 3 or more channels', icon='error')
            return
        procarr = self.ndarray.copy()
        procarr[:,:,0] = self.ndarray[:,:,2]
        procarr[:,:,1] = self.ndarray[:,:,1]
        procarr[:,:,2] = self.ndarray[:,:,0]
        self.show_array(procarr)


    def toMonochroom(self):
        array = self.ndarray

        if not array.ndim == 3:
            return

        dtype = array.dtype
        procarr = clip_array(array.mean(2), dtype)
        self.show_array(procarr)


    def toPhotoMonochroom(self):
        array = self.ndarray

        if not array.ndim == 3:
            return

        clip_low, clip_high = imconvert.integer_limits(array.dtype)
        mono = np.dot(array, [0.299, 0.587, 0.144])
        procarr = clip_array(mono, array.dtype)
        self.show_array(procarr)
        
    def is8bit(self):
        return self.ndarray.dtype in ['uint8', 'int8']
    
    def is16bit(self):
        return self.ndarray.dtype in ['uint16', 'int16']
        
    def to8bit(self):
        if not self.is16bit(): return
        self.show_array((self.ndarray >> 8).astype('uint8'))
        
    def to16bit(self):
        if not self.is8bit(): return
        self.show_array(self.ndarray.astype('uint16') << 8)

    def adjustLighting(self):
        """
        :param float offset:
        :param float gain:
        """
        with ActionArguments(self) as args:
            args['offset'] = -self.offset * 1.0
            args['gain'] = self.gain * 1.0

        if args.isNotSet():
            form = [('Offset', args['offset']),
                    ('Gain', args['gain'])]

            results = fedit(form, title='Adjust Lighting')
            if results is None: return
            offset, gain = results

        else:
            offset, gain = args['offset'], args['gain']

        #TO DO: use value mapping if possible
        array = self.ndarray
        procarr = clip_array(array * gain + offset, array.dtype)
        self.show_array(procarr)


    def adjustGamma(self):
        """
        :param float gamma:
        :param float upper:
        """
        with ActionArguments(self) as args:
            args['gamma'] = 1.0
            args['upper'] = 255

        if args.isNotSet():
            form = [('Gamma', args['gamma']),
                    ('Upper', args['upper'])]

            results = fedit(form, title='Adjust Gamma')
            if results is None: return
            gamma, upper = results

        else:
            gamma, upper = args['gamma'], args['upper']

        #TO DO: use value mapping if possible
        array = self.ndarray
        procarr = clip_array(np.power(array, gamma) * upper ** (1-gamma), array.dtype)
        self.show_array(procarr)


    ############################
    # Process Menu Connections

    def bayer_split_tiles(self):
        arr = self.ndarray
        blocks = []
        for x, y in [(0,0),(0,1),(1,0),(1,1)]:
            blocks.append(arr[y::2, x::2, ...])
        split = np.concatenate([
            np.concatenate([blocks[0], blocks[1]], axis=1),
            np.concatenate([blocks[2], blocks[3]], axis=1)])
        self.show_array(split)


    def colored_bayer(self):
        baypatns = ['RGGB', 'BGGR', 'GRBG', 'GBRG']
        form = [('Bayer Pattern', [1] + baypatns)]
        ind = fedit(form, title='Demosaic')[0]
        baypatn = baypatns[ind-1]

        procarr = bayer_split(self.ndarray, baypatn)
        self.show_array(procarr)


    def demosaic(self):
        baypatns = ['RGGB', 'BGGR', 'GRBG', 'GBRG']
        form = [('Bayer Pattern', [1] + baypatns)]
        ind = fedit(form, title='Demosaic')[0]
        baypatn = baypatns[ind-1]

        code = f"""\
        from gdesk.panels.imgview.demosaic import demosaicing_CFA_Bayer_bilinear
        procarr = demosaicing_CFA_Bayer_bilinear(gui.vs, '{baypatn}')
        gui.show(procarr)"""

        panel = gui.qapp.panels.select_or_new('console', None, 'child')
        panel.task.wait_process_ready()
        panel.exec_cmd(code)


    def makeBlueprint(self):
        with gui.qapp.waitCursor('making blueprint'):
            arr = self.ndarray

            if arr.ndim == 3:
                dtype = arr.dtype
                arr = arr.mean(2).astype(dtype)

            blueprint = make_thumbnail(arr)
            gui.img.new()
            gui.img.show(blueprint)


    def externalProcessDemo(self):
        panel = gui.qapp.panels.select_or_new('console', None, 'child')

        from .proxy import ImageGuiProxy

        def stage1_done(mode, error_code, result):
            gui.msgbox('Mirroring done')
            panel.task.call_func(ImageGuiProxy.high_pass_current_image, callback=stage2_done)

        def stage2_done(mode, error_code, result):
            gui.msgbox('Highpass filter done')

        panel.task.call_func(ImageGuiProxy.mirror_x, callback=stage1_done)


    def measureDistance(self):
        panel = gui.qapp.panels.select_or_new('console', None, 'child')

        panel.task.wait_process_ready()

        from .proxy import ImageGuiProxy

        def stage1_done(mode, error_code, result):
            pass

        panel.task.call_func(ImageGuiProxy.get_distance, callback=stage1_done)

    ############################
    # Analyse Menu Connections

    def horizontalSpectrogram(self):
        spectr_hori(self.ndarray)


    def verticalSpectrogram(self):
        spectr_vert(self.ndarray)

    #############################

    def show_array(self, array, zoomFitHist=False):
        self.refresh_offset_gain(array)
        self.contentChanged.emit(self.panid, zoomFitHist)


    def select(self):
        was_selected = super().select()
        if not was_selected:
            self.gainChanged.emit(self.panid)
            self.contentChanged.emit(self.panid, False)
        return was_selected


    def refresh_offset_gain(self, array=None, zoomFitHist=False):
        self.imviewer.imgdata.show_array(array, self.offset, self.gain, self.colormap, self.gamma)
        self.statuspanel.setOffsetGainInfo(self.offset, self.gain, self.gamma)
        self.gainChanged.emit(self.panid)
        self.imviewer.refresh()


    @property
    def ndarray(self):
        return self.sharray.ndarray


    @property
    def sharray(self):
        return self.imviewer.imgdata.sharray


class ImageViewer(ImageViewerBase):

    #contentChanged = Signal(int)
    #gainChanged = Signal(int)
    panelShortName = 'basic'
    userVisible = True

    def __init__(self, *args, **kwargs):
        #super().__init__(parent, panid, 'image')
        super().__init__(*args, **kwargs)

        self.imviewer = ImageViewerWidget(self)
        self.imviewer.roi.roiChanged.connect(self.passRoiChanged)
        self.imviewer.roi.get_context_menu = self.get_select_menu

        self.setCentralWidget(self.imviewer)
        self.imviewer.pickerPositionChanged.connect(self.set_info_xy_val)
        self.imviewer.zoomChanged.connect(self.statuspanel.set_zoom)
        self.imviewer.zoomPanChanged.connect(self.emitVisibleRegionChanged)


    def passRoiChanged(self):
        self.roiChanged.emit(self.panid)


    def emitVisibleRegionChanged(self):
        if self.imviewer.zoombind:
            self.visibleRegionChanged.emit(*self.imviewer.visibleRegion(normalized=True, clip_square=True), False, False, self.imviewer.zoomValue)
        else:
            self.visibleRegionChanged.emit(*self.imviewer.visibleRegion(normalized=True, clip_square=True), False, False, 0.0)


class ImageProfileWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)

        self.imviewer = ImageViewerWidget(self)

        self.profBtn = QtWidgets.QPushButton(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'diagramm.png')), None, self)
        self.profBtn.setFixedHeight(20)
        self.profBtn.setFixedWidth(20)
        self.profBtn.clicked.connect(self.toggleProfileVisible)
        self.rowPanel = ProfilerPanel(self, 'x', self.imviewer)
        self.colPanel = ProfilerPanel(self, 'y', self.imviewer)

        self.gridsplit = GridSplitter(None)


        self.imviewer.zoomPanChanged.connect(self.colPanel.zoomToImage)
        self.imviewer.zoomPanChanged.connect(self.rowPanel.zoomToImage)

        self.gridsplit.addWidget(self.rowPanel, 0, 1)
        self.gridsplit.addWidget(self.colPanel, 1, 0)
        self.gridsplit.addWidget(self.imviewer, 1, 1)

        self.cornerLayout = QtWidgets.QGridLayout()
        self.cornerLayout.addWidget(self.profBtn, 0, 0, alignment=Qt.AlignRight | Qt.AlignBottom)
        self.gridsplit.addLayout(self.cornerLayout, 0, 0)

        self.setLayout(self.gridsplit)

        self.profilesVisible = False


    def toggleProfileVisible(self):
        self.profilesVisible = not self.profilesVisible


    def showOnlyRuler(self):
        self.rowPanel.showOnlyRuler()
        self.colPanel.showOnlyRuler()
        self._profilesVisible = False

        gui.qapp.processEvents()
        self.refresh_profile_views()


    def showProfiles(self):
        self.rowPanel.setMinimumHeight(20)
        self.rowPanel.setMaximumHeight(500)
        self.colPanel.setMinimumWidth(20)
        self.colPanel.setMaximumWidth(500)

        strow = self.gridsplit.getRowStretches()
        stcol = self.gridsplit.getColumnStretches()
        rowspan = strow[0]+ strow[1]
        colspan = stcol[0] + stcol[1]
        target = rowspan // 5
        self.gridsplit.setRowStretches((target,rowspan-target))
        self.gridsplit.setColumnStretches((target,colspan-target))
        self.colPanel.showAll()
        self.rowPanel.showAll()

        self._profilesVisible = True
        self.drawMeanProfile()

        gui.qapp.processEvents()
        self.refresh_profile_views()


    def drawMeanProfile(self):
        arr = self.imviewer.imgdata.sharray.ndarray

        if arr.ndim > 2:
            arr = arr.mean(2)

        rowProfile = arr.mean(0)
        colProfile = arr.mean(1)

        self.rowPanel.drawMeanProfile(np.arange(len(rowProfile)), rowProfile)
        self.colPanel.drawMeanProfile(np.arange(len(colProfile)), colProfile)

        self.refresh_profile_views()


    def set_profiles_visible(self, value):
        if value:
            self.showProfiles()
        else:
            self.showOnlyRuler()

    profilesVisible = property(lambda self: self._profilesVisible, set_profiles_visible)

    def refresh_profile_views(self):
        self.colPanel.zoomToImage()
        self.rowPanel.zoomToImage()

        if self.colPanel.view.auto_zoom:
            self.colPanel.zoomAuto()
        self.colPanel.view.refresh()
        if self.rowPanel.view.auto_zoom:
            self.rowPanel.zoomAuto()
        self.rowPanel.view.refresh()


class ImageProfilePanel(ImageViewerBase):
    panelShortName = 'image-profile'
    userVisible = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.imgprof = ImageProfileWidget(self)
        self.setCentralWidget(self.imgprof)

        self.imviewer.pickerPositionChanged.connect(self.set_info_xy_val)
        self.imviewer.zoomChanged.connect(self.statuspanel.set_zoom)
        self.imviewer.zoomPanChanged.connect(self.emitVisibleRegionChanged)
        self.imviewer.roi.roiChanged.connect(self.passRoiChanged)
        self.imviewer.roi.get_context_menu = self.get_select_menu

        self.addMenuItem(self.viewMenu, 'Show/Hide Profiles'    , self.showHideProfiles,
            checkcall=lambda: self.imgprof.profilesVisible,
            statusTip="Show or Hide the image column and row profiles")


    def emitVisibleRegionChanged(self):
        if self.imviewer.zoombind:
            self.visibleRegionChanged.emit(*self.imviewer.visibleRegion(normalized=True, clip_square=True), False, False, self.imviewer.zoomValue)
        else:
            self.visibleRegionChanged.emit(*self.imviewer.visibleRegion(normalized=True, clip_square=True), False, False, 0.0)


    def changeVisibleRegion(self, x, y, w, h, zoomSnap, emit, zoomValue):
        self.imgprof.imviewer.zoomNormalized(x, y, w, h, zoomSnap, emit, zoomValue)
        self.imgprof.colPanel.zoomToImage()
        self.imgprof.rowPanel.zoomToImage()
        self.imviewer.roi.recalcGeometry()


    def passRoiChanged(self):
        self.roiChanged.emit(self.panid)


    def show_array(self, array, zoomFitHist=False):
        super().show_array(array, zoomFitHist)
        if self.imgprof.profilesVisible:
            self.imgprof.drawMeanProfile()
        else:
            self.imgprof.refresh_profile_views()


    @property
    def imviewer(self):
        return self.imgprof.imviewer


    def showHideProfiles(self):
        self.imgprof.profilesVisible = not self.imgprof.profilesVisible




import logging
import time
import sys, os
import threading
import queue
import socket
import json

import zmq

from . import conf
from .conf import config
from .gui_proxy import gui

logger = logging.getLogger(__name__)
context = zmq.Context()

class CommandServer(object):

    def __init__(self, shell):
        self.shell = shell
        self.server_thread = None
        self.socket_loop = False
        self.queue_loop = False
        self.cmd_queue = queue.Queue()

    def start(self, qapp):
        self.socket = context.socket(zmq.REP)

        min_port = config.get('zmq_watcher_min_port', 5550) 
        max_port = config.get('zmq_watcher_max_port', 5560)                
        
        try:
            self.port = self.socket.bind_to_random_port('tcp://*',
                min_port=min_port, max_port=max_port, max_tries=100)        
            logger.info(self.host_info())
            
        except zmq.error.ZMQError as err:
            if err.strerror == 'Address in use':
                self.socket.close()
                logger.info(f'Watcher port {self.port} is already in use')
                return False
            else:
                raise        
                
        with open(self.shell.logdir.logpath / 'cmdserver.json', 'w') as fp:
            info = {'port': self.port}
            json.dump(info, fp)
        
        self.socket_loop = True
        self.server_thread = threading.Thread(target=self.recv_socket_socket_loop, args=(qapp.handover,))
        self.server_thread.setDaemon(True)
        self.server_thread.start()
        return True
        
    def host_info(self):
        hostname_ex, aliaslist, ipaddrlist = socket.gethostbyname_ex(socket.gethostname())
        message = []
        message.append(f'Watcher port : {self.port}')
        message.append(f'Hostname     : {hostname_ex}')
        message.append(f'Aliases      : {aliaslist}')
        message.append(f'IP addresses : {ipaddrlist}')
        return '\n'.join(message)
        
    def start_queue_loop(self, qapp):        
        self.queue_loop = True
        self.cmd_queue_thread = threading.Thread(target=self.recv_queue_socket_loop, args=(qapp.handover,))
        self.cmd_queue_thread.setDaemon(True)
        self.cmd_queue_thread.start()        
        
    def stop(self):
        self.socket_loop = False
        self.socket = context.socket(zmq.REQ)
        self.socket.connect(f"tcp://localhost:{self.port}")
        self.socket.send_pyobj('close')
        self.socket.close()
        
    @staticmethod
    def open_images(*image_paths):
        for image_path in image_paths:
            gui.img.open(image_path)
            
    @staticmethod        
    def connect_process(cqs_config=None):
        from .tasks import ZmqQueues
        print(f'cqs_config:{cqs_config}')
        
        if cqs_config is None:        
            cqs = conf.config_objects['cqs']
        else:
            cqs = ZmqQueues.from_json(cqs_config)
            cqs.setup_as_client()
            
        gui.qapp.panels['console'][0].get_container().window().hide()
        gui.qapp.panels.select_or_new('console', None, 'child', args=(cqs,))
        
        return cqs_config
        
    @staticmethod        
    def connect_zmq_process(cqs_config=None):
        from .tasks import ZmqQueues       
        
        if cqs_config is None:       
            cqs = ZmqQueues()
            cqs.setup_as_server()
        else:
            cqs = ZmqQueues.from_json(cqs_config)
            cqs.setup_as_client(cqs.hostname_ex)
            
        gui.qapp.panels['console'][0].get_container().window().hide()
        gui.qapp.panels.select_or_new('console', None, 'child', args=(cqs,))
        
        return cqs.to_json()        
        
    @staticmethod        
    def execute_file(init_file, console_id):                
        gui.console.execute_file(init_file, console_id)     

    @staticmethod        
    def execute_code(init_code, console_id):        
        from .. import gui
        gui.console.execute_code(init_code, console_id)          

    @staticmethod
    def start_kernel(kerneltype='ipykernel', connectfile='', rundir=None, child=True, threaded=False):
        from .. import gui
        from ..external import jupyter
        
        if child:
            gui.console.child(jupyter.start_kernel, kerneltype, connectfile, rundir, threaded)
        else:
            pass
                      
    def recv_socket_socket_loop(self, handover):
        while self.socket_loop:
            request = self.socket.recv_pyobj()
            
            if request == 'close':
                break
                
            else:
                cmd, args = request['cmd'], request['args']
                        
            answer = self.execute_command(handover, cmd, args)
            self.socket.send_pyobj(answer or request)

        self.socket.close()
        logger.info(f'Watcher stoppped')
        
    def recv_queue_socket_loop(self, handover):
        while self.queue_loop:
            request = self.cmd_queue.get()
            
            if request == 'close':
                break
                
            else:
                cmd, args = request['cmd'], request['args']
                
            answer = self.execute_command(handover, cmd, args)                              
                        
        
    def execute_command(self, handover, cmd, args):
        if cmd is None:
            return
            
        elif cmd == 'open_images':            
            return handover.send(True, CommandServer.open_images, *args)
            
        elif cmd == 'start_kernel':
            return handover.send(True, CommandServer.start_kernel, *args)
            
        elif cmd == 'connect_process':
            return handover.send(True, CommandServer.connect_process, *args)
            
        elif cmd == 'connect_zmq_process':
            return handover.send(True, CommandServer.connect_zmq_process, *args)            
            
        elif cmd == 'execute_file':
            return handover.send(True, CommandServer.execute_file, *args)
            
        elif cmd == 'execute_code':
            return handover.send(True, CommandServer.execute_code, *args)            
        
class CommandClient(object):

    def __init__(self, port=None, host='localhost'):
        self.timeout = 10000
        self.retries = 1
        self.port = port or config.get('watcher_port', 5998)
        self.host = host       
    
    def send(self, message, timeout=None, retries=None):
        timeout = timeout or self.timeout
        retries_left = retries or self.retries
        server_endpoint = f"tcp://{self.host}:{self.port}"
        logging.info(f"Connecting to server {server_endpoint}...")
        
        self.socket = context.socket(zmq.REQ)
        
        try:
            self.socket.connect(server_endpoint)
        except:
            raise

        request = message
        logging.debug(f"Sending {request}")
        self.socket.send_pyobj(request)
        
        while True:
            if (self.socket.poll(timeout) & zmq.POLLIN) != 0:
                reply = self.socket.recv_pyobj()
                return reply

            retries_left -= 1
            logging.warning("No response from server")
            # Socket is confused. Close and remove it.
            self.socket.setsockopt(zmq.LINGER, 0)
            self.socket.close()
            if retries_left == 0:
                logging.error("Server seems to be offline, abandoning")
                break

            logging.info("Reconnecting to server...")
            # Create new connection
            self.socket = context.socket(zmq.REQ)
            self.socket.connect(server_endpoint)
            logging.info("Resending (%s)", request)
            self.socket.send_pyobj(request)

        self.socket.close()


import numpy as np
from qtpy import QtCore, QtGui, QtWidgets

QtSignal = QtCore.Signal

class ItemSignal(object):
    def __init__(self):
        self.func = None
    
    def connect(self, func):
        self.func = func
        
    def emit(self, *args):
        if not self.func is None:
            self.func(*args)            
            
class VectorCurve(QtWidgets.QGraphicsPathItem):
    
    def __init__(self, path, xvector, yvector):
        super().__init__(path)
        self.xvector = xvector
        self.yvector = yvector


def createCurve(x, y, color=None, z=0):
    if color == None:
        pen = QtGui.QPen(QtCore.Qt.black, 0, QtCore.Qt.SolidLine)
        brush = QtGui.QBrush(QtGui.QColor(0,0,0,100))
        
    else:
        pen = QtGui.QPen(color, 0, QtCore.Qt.SolidLine)
        R,G,B,A = color.toTuple()
        brush = QtGui.QBrush(QtGui.QColor(R,G,B,50))
        
    #first create a Path
    path = QtGui.QPainterPath()
    
    path.moveTo(x[0], 0)    
    path.lineTo(x[0], y[0])
    
    for i in range(1, len(y)):
        path.lineTo(x[i], y[i]) 
        
    path.lineTo(x[-1], 0)    
            
    #transform the Path to a PathItem                                    
    #curve = QtWidgets.QGraphicsPathItem(path)
    curve = VectorCurve(path, np.array(x), np.array(y))
    if z != 0:
        curve.setZValue(z)
        
    curve.setPen(pen)
    curve.setBrush(brush)
    
    return curve
    
    
class LabelItem(QtWidgets.QGraphicsPolygonItem):
    
    def __init__(self, text='', color=QtGui.QColor(0,0,0), parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        
        self.makePolygon(40)
        self.setPen(QtGui.QPen(color))
        self.setBrush(QtGui.QColor(240, 240, 240))
        
        self.setFlag(QtWidgets.QGraphicsItem.ItemIgnoresTransformations, True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, True)
        #self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, False)
        
        self.label = QtWidgets.QGraphicsTextItem('', self)
        self.label.setFont(QtGui.QFont('Arial', 8))
        self.label.setPos(-1, 0)
        self.updateText(text)            
        
    def makePolygon(self, box_width):
        polygon = QtGui.QPolygonF()        
        polygon.append(QtCore.QPointF(0, 0))
        polygon.append(QtCore.QPointF(box_width / 2, 5))
        polygon.append(QtCore.QPointF(box_width / 2, 20))
        polygon.append(QtCore.QPointF(-box_width / 2, 20))
        polygon.append(QtCore.QPointF(-box_width / 2, 5))
        polygon.append(QtCore.QPointF(0, 0))
        self.setPolygon(polygon)        
        
    def updateText(self, text):
        self.label.setPlainText(text)
        self.label.setPos(- self.label.boundingRect().width() / 2, 2)    
        self.makePolygon(self.label.boundingRect().width())
        
    def mouseMoveEvent(self, e):
        self.parentItem().mouseMoveEvent(e)
        
    def mouseReleaseEvent(self, event):
        self.parentItem().mouseReleaseEvent(event)
        super().mouseReleaseEvent(event)
        
        
class YLabelItem(QtWidgets.QGraphicsPolygonItem):
    
    def __init__(self, text='', color=QtGui.QColor(0,0,0), parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
                
        self.setPen(QtGui.QPen(color))
        self.setBrush(QtGui.QColor(240, 240, 240))        
        self.setFlag(QtWidgets.QGraphicsItem.ItemIgnoresTransformations, True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, True)
        
        self.label = QtWidgets.QGraphicsTextItem('', self)
        self.label.setFont(QtGui.QFont('Arial', 8))
                        
        self.offset = 0
        self.updateText(text)
        
    def update_offset(self, offset):
        self.offset = offset
        self.makePolygon()
        
    def makePolygon(self):
        box_width = self.text_width
        offset = self.offset
        polygon = QtGui.QPolygonF()        
        polygon.append(QtCore.QPointF(0, 0))
        polygon.append(QtCore.QPointF(5, -5 + offset))
        polygon.append(QtCore.QPointF(5, -20 + offset))
        polygon.append(QtCore.QPointF(5 + box_width, -20 + offset))
        polygon.append(QtCore.QPointF(5 + box_width, -5 + offset))
        polygon.append(QtCore.QPointF(5, -5+offset))
        polygon.append(QtCore.QPointF(0, 0))
        self.label.setPos(5, -23 + offset)
        self.setPolygon(polygon)        
        
    def updateText(self, text):        
        self.label.setPlainText(text) 
        self.text_width = self.label.boundingRect().width()
        self.makePolygon()
        
    def mouseMoveEvent(self, e):
        self.parentItem().mouseMoveEvent(e)      

    def sortkey(self):
        return self.pos().y()
    
    
class Indicator(QtWidgets.QGraphicsPolygonItem):
    
    def __init__(self, color = QtCore.Qt.blue, text = None, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        self.mouse_released = ItemSignal()
        
        self.setFlag(QtWidgets.QGraphicsItem.ItemIgnoresTransformations, True)        
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, False)
        
        self.setPen(QtGui.QPen(color))
                
        polygon = QtGui.QPolygonF()
        
        self.text = text
        
        if text is None:            
            self.setBrush(color)
            polygon.append(QtCore.QPointF(0, 0))
            polygon.append(QtCore.QPointF(7, 10))
            polygon.append(QtCore.QPointF(-7, 10))
            polygon.append(QtCore.QPointF(0, 0))
            polygon.append(QtCore.QPointF(0, -2160))
            self.setPolygon(polygon)
            self.label = None
        else:
            polygon.append(QtCore.QPointF(0, 0))
            polygon.append(QtCore.QPointF(0, -2160))
            polygon.append(QtCore.QPointF(0, 0))
            self.setPolygon(polygon)
            self.label = LabelItem(self.text, color, self, scene)
            #self.addItem(self.label)
            
        self.ylabels = []
        
    def attach_curves(self, curves):
        self.curves = curves
        
    def set_ylabel_count(self, count):
        for i in range(len(self.ylabels) - count):
            ylabel = self.ylabels.pop(0)
            self.scene().removeItem(ylabel)
        
        for i in range(count - len(self.ylabels)):
            ylabel = YLabelItem('test', parent = self)
            self.ylabels.append(ylabel)                                
        
    def updates_ylabels(self, x=None):        
        x = self.scenePos().x() if x is None else x
        view = self.scene().views()[0]
        self.set_ylabel_count(len(self.curves))
        for curve, ylabel in zip(self.curves.values(), self.ylabels):
            yval = np.interp(x, curve.xvector, curve.yvector,0,0)                                
            yscale = view.scale[1]
            ypos = (yval - self.scenePos().y()) * yscale
            if ypos > 0:
                ypos = 0
            elif ypos < (-view.height()+23):
                ypos = -view.height()+23
            ylabel.setPos(0, ypos)
            ylabel.updateText("%0.4g" % yval)                    
            ylabel.setPen(curve.pen())
        self.declutter_ylabels(-view.height()+22)

    def declutter_ylabels(self, ymin=-4000, ymax=0):
        self.ylabels = sorted(self.ylabels, key = YLabelItem.sortkey)
        prior_bottom = ymin
        for ylabel in self.ylabels:
            ypos = ylabel.pos().y()
            if (ypos - prior_bottom) < 21:
                offset = abs(21 - (ypos - prior_bottom))
            else:
                offset = 0
            ylabel.update_offset(offset)
            prior_bottom = ypos -5 + offset
        
    def mouseReleaseEvent(self, event):        
        self.mouse_released.emit(self.pos().x())
        super().mouseReleaseEvent(event)        
                
    def mouseMoveEvent(self, event):        
        x = event.scenePos().x()
        if (not self.label is None) and ('%' in self.text):            
            self.label.updateText(self.text % x)
        
        self.setPos(x, 0)            
        self.updates_ylabels(x)
        
        self.scene().moving_indicators = True


class Grid(QtWidgets.QGraphicsItem):

    def __init__(self, direction, parent=None, scene=None):
        super().__init__(parent=parent, scene=scene)
        pens = []   
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))
        pens.append(QtGui.QPen(QtGui.QColor(159,159,159), 0, QtCore.Qt.SolidLine))
        pens.append(QtGui.QPen(QtGui.QColor(191,191,191), 0, QtCore.Qt.DashLine))
        pens.append(QtGui.QPen(QtGui.QColor(223,223,223), 0, QtCore.Qt.DotLine))        
        
        self.pens = pens
        self.grid_items = []       
        self.direction = direction
        
    def boundingRect(self):
        return QtCore.QRectF(-2, -2, 4, 4)

    def paint(self, painter, option, widget):
        pass
        #painter.drawRoundedRect(-10, -10, 20, 20, 5, 5)        
               
    def attach_rulers(self, x_ruler, y_ruler):
        self.x_ruler = x_ruler
        self.y_ruler = y_ruler
        self.update_grid()        
        
    def update_grid(self):
        for path_item in self.grid_items:
            self.scene().removeItem(path_item)
            
        self.grid_items = []            
            
        pens = self.pens             
        paths = []      
        view = self.scene().views()[0]                
                           
        for thicklevel in range(3):
            paths.append(QtGui.QPainterPath())
            for i in self.x_ruler.thicks[thicklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(i, self.y_ruler.start)
                    paths[-1].lineTo(i, self.y_ruler.stop)               
                else:
                    paths[-1].moveTo(self.x_ruler.start, i)
                    paths[-1].lineTo(self.x_ruler.stop, i) 
                               
        for thicklevel in range(3):        
            paths.append(QtGui.QPainterPath())  
            for i in self.y_ruler.thicks[thicklevel][1]:
                if self.direction == 0:
                    paths[-1].moveTo(self.x_ruler.start, i)
                    paths[-1].lineTo(self.x_ruler.stop, i)
                else:
                    paths[-1].moveTo(i, self.y_ruler.start)
                    paths[-1].lineTo(i, self.y_ruler.stop)
                
        for i in range(len(paths)):
            path_item = QtWidgets.QGraphicsPathItem(paths[i], parent=self)
            path_item.setPen(pens[i])
            path_item.setZValue(-2)
            self.grid_items.append(path_item)         
        